import pytest

import autofit as af
from test_autofit import mock

names = ["one", "two", "three"]


@pytest.fixture(
    name="collection"
)
def make_collection():
    return af.Collection({
        name: af.PriorModel(
            mock.MockClassx2
        )
        for name in names
    })


def test_prior_count(collection):
    assert collection.prior_count == 6


@pytest.mark.parametrize(
    "name",
    names
)
def test_children(collection, name):
    assert getattr(collection, name).prior_count == 2


def test_replace(collection):
    collection.one = af.PriorModel(
        mock.MockClassx4
    )

    assert collection.one.prior_count == 4
    assert collection.prior_count == 8



from howtofit.simulators.chapter_1 import profiles

import numpy as np

# %%
"""
__Gaussian X2 + Exponential x1__

Create a model instance of both Gaussians and the Exponential.
"""

# %%
gaussian_0 = profiles.Gaussian(centre=50.0, intensity=25.0, sigma=10.0)
gaussian_1 = profiles.Gaussian(centre=20.0, intensity=30.0, sigma=5.0)
exponential = profiles.Exponential(centre=70.0, intensity=40.0, rate=0.005)

# %%
"""
Specify the number of pixels used to create the xvalues on which the 1D line of the profile is generated using and
thus defining the number of data-points in our data.
"""

# %%
pixels = 100
xvalues = np.arange(pixels)

# %%
"""
Evaluate both Gaussian model instances and Exponential model instance at every xvalues to create their model line
profiles and sum them together to create the overall model profile.
"""

# %%
model_line = (
    gaussian_0.profile_from_xvalues(xvalues=xvalues)
    + gaussian_1.profile_from_xvalues(xvalues=xvalues)
    + exponential.profile_from_xvalues(xvalues=xvalues)
)

# %%
"""
Determine the noise (at a specified signal to noise level) in every pixel of our model profile.
"""

# %%
signal_to_noise_ratio = 25.0
noise = np.random.normal(0.0, 1.0 / signal_to_noise_ratio, pixels)

# %%
"""
Add this noise to the model line to create the line data that is fitted, using the signal-to-noise ratio to compute
noise-map of our data which is required when evaluating the chi-squared value of the likelihood.
"""

# %%
data = model_line + noise
noise_map = (1.0 / signal_to_noise_ratio) * np.ones(pixels)



import autofit as af

"""
The Result class stores the results of a non-linear search's, which are inherited from autofit's Result object. In
tutorial 3 we discussed some of the results that it contains and we will discuss them further in tutorial 8.

By using the PyAutoFit phase API, we are able to extend the Result object returned by a phase's non-linear search to
include model-specific properties of the fit. Below, we include method that return the model data and fit of the
maximum log likelihood model, which we use in the main tutorial script to plot the results.
"""


class Result(af.Result):
    def __init__(self, samples, previous_model, search, analysis):
        """
        The results of a non-linear search performed by a phase.

        Parameters
        ----------
        samples : af.Samples
            A class containing the samples of the non-linear search, including methods to get the maximum log
            likelihood model, errors, etc.
        analysis : Analysis
            The Analysis class used by this model-fit to fit the model to the data.
        """
        super().__init__(samples=samples, previous_model=previous_model, search=search)
        self.analysis = analysis

    @property
    def max_log_likelihood_model_data(self):
        return self.analysis.model_data_from_instance(
            instance=self.samples.max_log_likelihood_instance
        )

    @property
    def max_log_likelihood_fit(self):
        return self.analysis.fit_from_model_data(
            model_data=self.max_log_likelihood_model_data
        )



from howtofit.chapter_1_introduction.tutorial_8_aggregator.src.plot import line_plots

"""The 'dataset_plots.py' module is unchanged from the previous tutorial."""


def data(dataset, output_path=None, output_filename=None, output_format="show"):
    """Plot the data values of a Line dataset.

    Parameters
    -----------
    Line : dataset.Line
        The observed Line dataset whose data is plotted.
    """
    line_plots.line(
        xvalues=dataset.xvalues,
        line=dataset.data,
        ylabel="Data Values",
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )


def noise_map(dataset, output_path=None, output_filename=None, output_format="show"):
    """Plot the noise-map of a Line dataset.

    Parameters
    -----------
    Line : dataset.Line
        The observed Line dataset whose data is plotted.
    """
    line_plots.line(
        xvalues=dataset.xvalues,
        line=dataset.noise_map,
        ylabel="Noise-Map",
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )



import numpy as np

"""The Dataset class of the 'dataset.py' module is unchanged from the previous tutorial."""


class Dataset:
    def __init__(self, data, noise_map):
        """A class containing the data and noise-map of a 1D line dataset.

        Parameters
        ----------
        data : np.ndarray
            The array of the data, in arbitrary units.
        noise_map : np.ndarray
            An array describing the RMS standard deviation error in each data pixel, in arbitrary units.
        """
        self.data = data
        self.noise_map = noise_map

    @property
    def xvalues(self):
        return np.arange(self.data.shape[0])


"""
Here, we create masked dataset that is fitted by our phase. This class takes an unmasked dataset (e.g. an image and
noise-map) and applies a mask to them, such that all entries where the mask is True are omitted from the fit and
log_likelihood calution.

This could be done using NumPy masked array functionality, by for simplicity we will simply set all masked entries
to zero instead (and not included them in the fit as seen in the 'fit.py' module).

If your model-fitting problem requires masking you'll want a module something very similar to this one!
"""


class MaskedDataset:
    def __init__(self, dataset, mask):
        """
        A masked dataset, which is an image, noise-map and mask.

        Parameters
        ----------
        dataset: im.Dataset
            The dataset (the image, noise-map, etc.)
        mask: msk.Mask
            The 1D mask that is applied to the dataset.
        """

        """We store the unmasked dataset in the masked-dataset, incase we need it for anything."""
        self.dataset = dataset

        self.mask = mask

        """We apply the mask, setting all entries where the mask is True to zero."""
        self.data = dataset.data * np.invert(mask)

        """Same for the noise-map"""
        self.noise_map = dataset.noise_map * np.invert(mask)

    @property
    def xvalues(self):
        return np.arange(self.data.shape[0])

    def signal_to_noise_map(self):
        return self.data / self.noise_map



from abc import ABC, abstractmethod

from autofit.mapper.prior.arithmetic import ArithmeticMixin
from autofit.tools.pipeline import ResultsCollection


class AbstractPromiseResult(ABC):
    def __init__(self, *result_path):
        self.result_path = result_path

    @property
    @abstractmethod
    def model(self):
        """
        A promise for an object in the model result. This might be a prior or prior model.
        """

    @property
    @abstractmethod
    def instance(self):
        """
        A promise for an object in the best fit result. This must be an instance or instance.
        """

    def model_absolute(self, a) -> "AbstractPromise":
        """
        Parameters
        ----------
        a
            The absolute width of gaussian priors

        Returns
        -------
        A promise for a prior or model with absolute gaussian widths
        """
        model = self.model
        model.absolute = a
        return model

    def model_relative(self, r) -> "AbstractPromise":
        """
        Parameters
        ----------
        r
            The relative width of gaussian priors

        Returns
        -------
        A promise for a prior or model with relative gaussian widths
        """
        model = self.model
        model.relative = r
        return model

    @abstractmethod
    def __getattr__(self, item):
        """
        Used to manage results paths
        """

    def __call__(self, *args, **kwargs):
        pass


class LastPromiseResult(AbstractPromiseResult):
    """
    A PromiseResult that does not require a phase. Refers to the latest Result in the collection with an object for
    the path specified in a promise.
    """

    def __init__(self, *result_path, index=0):
        super().__init__(*result_path)
        self._index = index

    def __getitem__(self, item: int) -> "LastPromiseResult":
        """
        LastPromiseResults can be indexed.

        An index of 0 gives a LPR corresponding to the latest Results.
        An index of -1 gives a LPR corresponding the the Results before last.
        etc.
        """
        if item > 0:
            raise IndexError("last only accepts negative indices")
        return LastPromiseResult(*self.result_path, index=item)

    @property
    def model(self):
        """
        A promise for an object in the model result. This might be a prior or prior model.
        """
        return LastPromise(result_path=self.result_path, index=self._index)

    @property
    def instance(self):
        """
        A promise for an object in the best fit result. This must be an instance or instance.
        """
        return LastPromise(
            result_path=self.result_path, is_instance=True, index=self._index
        )

    def __getattr__(self, item):
        if item == "_index":
            return super().__getattr__(item)
        return LastPromiseResult(*self.result_path, item)


class PromiseResult(AbstractPromiseResult):
    def __init__(self, phase, *result_path, assert_exists=True):
        """
        A wrapper for a phase that facilitates the generation of priors by consuming user defined paths.

        Parameters
        ----------
        phase
            A phase from which a promise is generated
        result_path
            The path through the results objects to the correct result. This is used for phase extensions when
            a result object can have child result objects
        assert_exists
            If this is true then an AttributeError is thrown if there is no object with the given path in the
            model of the origin phase
        """
        super().__init__(*result_path)
        self._phase = phase
        self.assert_exists = assert_exists

    def __getstate__(self):
        return {}

    @property
    def model(self):
        """
        A promise for an object in the model result. This might be a prior or prior model.
        """
        return Promise(
            self._phase,
            result_path=self.result_path,
            assert_exists=self.assert_exists
        )

    @property
    def instance(self):
        """
        A promise for an object in the best fit result. This must be an instance or instance.
        """
        return Promise(
            self._phase,
            result_path=self.result_path,
            is_instance=True,
            assert_exists=self.assert_exists,
        )

    def __getattr__(self, item):
        return PromiseResult(
            self._phase,
            *self.result_path,
            item,
            assert_exists=False
        )


class AbstractPromise(ABC, ArithmeticMixin):
    def __init__(
            self,
            *path,
            result_path,
            is_instance=False,
            absolute=None,
            relative=None,
            is_optional=False,
    ):
        """
        Place holder for an object in the object hierarchy. This is replaced at runtime by a prior, prior
        model, instance or instance

        Parameters
        ----------
        path
            The path to the promised object. e.g. if a phase has a model galaxies.lens.phi then the path
            will be ("galaxies", "lens", "phi")
        result_path
            The path through the result collection to the result object required. This is used for hyper phases
            where a result object can have child result objects for each phase extension.
        is_instance
            True if the promised object belongs to the instance result object. Otherwise it belongs to the
            model.
        """
        self.path = path
        self.is_instance = is_instance
        self.result_path = result_path
        self.absolute = absolute
        self.relative = relative
        self.is_optional = is_optional

    @property
    def attribute_name(self) -> str:
        """
        The name of the attribute which was called to create this prior.
        """
        return "instance" if self.is_instance else "model"

    @property
    def path_string(self):
        """
        A string expressing the path of attributes called to create this prior.
        """
        return ".".join(map(str, self.path))

    def __str__(self):
        return f"result.{self.attribute_name}.{self.path_string}"

    @property
    def settings(self):
        return dict(
            is_instance=self.is_instance,
            absolute=self.absolute,
            relative=self.relative,
            is_optional=self.is_optional,
        )

    def __call__(self, *args, **kwargs):
        pass

    @abstractmethod
    def populate(self, results_collection):
        """
        Find the object that this promise corresponds to be getting results for a particular phase and then
        traversing those results using the path.

        Parameters
        ----------
        results_collection
            A collection of results from previous phases

        Returns
        -------
        obj
            The promised prior, prior model, instance or instance
        """

    def _populate_from_results(self, results):
        for item in self.result_path:
            try:
                results = getattr(results, item)
            except AttributeError as e:
                if self.is_optional:
                    return None
                raise e
        if self.absolute is not None:
            model = results.model_absolute(self.absolute)
        elif self.relative is not None:
            model = results.model_relative(self.relative)
        elif self.is_instance:
            model = results.instance
        else:
            model = results.model

        try:
            return model.object_for_path(self.path)
        except AttributeError as e:
            if self.is_optional:
                return None
            raise e


class Promise(AbstractPromise):
    def __init__(
            self,
            phase,
            *path,
            result_path,
            is_instance=False,
            assert_exists=True,
            relative=None,
            absolute=None,
            is_optional=False,
    ):
        """
        Place holder for an object in the object hierarchy. This is replaced at runtime by a prior, prior
        model, instance or instance

        Parameters
        ----------
        phase
            The phase that holds or creates the promised object
        path
            The path to the promised object. e.g. if a phase has a model galaxies.lens.phi then the path
            will be ("galaxies", "lens", "phi")
        result_path
            The path through the result collection to the result object required. This is used for hyper phases
            where a result object can have child result objects for each phase extension.
        is_instance
            True if the promised object belongs to the instance result object
        assert_exists
            If this is true then an exception is raised if an object is not defined in the addressed phase's
            model. Hyper phases are a bit trickier so no assertion is made.
        """
        super().__init__(
            *path,
            result_path=result_path,
            is_instance=is_instance,
            absolute=absolute,
            relative=relative,
            is_optional=is_optional,
        )
        self._phase = phase
        self.assert_exists = assert_exists
        if assert_exists:
            phase.model.object_for_path(path)

    def __str__(self):
        return f"{self._phase.phase_name}.{super().__str__()}"

    def __getstate__(self):
        return {}

    def __getattr__(self, item):
        if item in (
                "_phase",
                "path",
                "is_instance",
                "_populate_from_results",
                "optional",
                "is_optional",
        ):
            return super().__getattribute__(item)
        return Promise(
            self._phase,
            *self.path,
            item,
            result_path=self.result_path,
            assert_exists=self.assert_exists,
            **self.settings,
        )

    def __getitem__(self, item):
        return Promise(
            self._phase,
            *self.path,
            item,
            result_path=self.result_path,
            assert_exists=self.assert_exists,
            **self.settings,
        )

    @property
    def optional(self) -> "Promise":
        """
        Make this promise evaluate to None if a corresponding float or prior
        does not exist at evaluation time. This is instead of raising an
        AttributeError.
        """
        return Promise(
            self._phase,
            *self.path,
            result_path=self.result_path,
            assert_exists=False,
            **{
                key: value
                for key, value in self.settings.items()
                if key != "is_optional"
            },
            is_optional=True,
        )

    def populate(self, results_collection):
        """
        Find the object that this promise corresponds to be getting results for a particular phase and then
        traversing those results using the path.

        Parameters
        ----------
        results_collection
            A collection of results from previous phases

        Returns
        -------
        obj
            The promised prior, prior model, instance or instance
        """
        results = results_collection.from_phase(
            self._phase.phase_name
        )
        return self._populate_from_results(results)


class LastPromise(AbstractPromise):
    """
    A promise that searches the results collection to find the latest result that contains an object with the
    specified path
    """

    def __init__(
            self,
            *path,
            result_path,
            is_instance=False,
            index=0,
            absolute=None,
            relative=None,
            is_optional=False,
    ):
        self._index = index
        super().__init__(
            *path,
            result_path=result_path,
            is_instance=is_instance,
            relative=relative,
            absolute=absolute,
            is_optional=is_optional,
        )

    def __getattr__(self, item):
        if item in (
                "phase",
                "path",
                "is_instance",
                "_populate_from_results",
                "optional",
                "is_optional",
        ):
            return super().__getattribute__(item)
        return LastPromise(
            *self.path,
            item,
            result_path=self.result_path,
            index=self._index,
            **self.settings,
        )

    @property
    def optional(self):
        return LastPromise(
            *self.path,
            result_path=self.result_path,
            index=self._index,
            **{
                key: value
                for key, value in self.settings.items()
                if key != "is_optional"
            },
            is_optional=True,
        )

    def populate(self, results_collection: ResultsCollection):
        """
        Recover the instance or model associated with this promise from the latest result in the results collection
        where a matching path is found.
        
        Parameters
        ----------
        results_collection
            A collection of results to be searched in reverse

        Returns
        -------
        A prior, model or instance.
        
        Raises
        ------
        AttributeError
            If no matching prior is found
        """
        for results in list(results_collection.reversed)[-self._index:]:
            try:
                return self._populate_from_results(results)
            except AttributeError:
                pass
        raise AttributeError(
            f"No attribute found with path {self.path} in previous phase"
        )


last = LastPromiseResult()



import numpy as np
import json

from autoconf import conf
from autofit.non_linear.initializer import InitializerPrior
from autofit.non_linear.abstract_search import NonLinearSearch
from autofit.non_linear.abstract_search import IntervalCounter
from autofit.non_linear.paths import Paths
from autofit.non_linear import samples as samp

from autofit import exc


class AbstractNest(NonLinearSearch):
    def __init__(
        self,
        paths=None,
        prior_passer=None,
        iterations_per_update=None,
        terminate_at_acceptance_ratio=None,
        acceptance_ratio_threshold=None,
        stagger_resampling_likelihood=None,
    ):
        """
        Abstract class of a nested sampling non-linear search (e.g. MultiNest, Dynesty).

        **PyAutoFit** allows a nested sampler to automatically terminate when the acceptance ratio falls below an input
        threshold value. When this occurs, all samples are accepted using the current maximum log likelihood value,
        irrespective of how well the model actually fits the data.

        This feature should be used for non-linear searches where the nested sampler gets 'stuck', for example because
        the log likelihood function is stochastic or varies rapidly over small scales in parameter space. The results of
        samples using this feature are not realiable (given the log likelihood is being manipulated to end the run), but
        they are still valid results for linking priors to a new phase and non-linear search.

        Parameters
        ----------
        paths : af.Paths
            Manages all paths, e.g. where the search outputs are stored, the samples, backups, etc.
        prior_passer : af.PriorPasser
            Controls how priors are passed from the results of this non-linear search to a subsequent non-linear search.
        terminate_at_acceptance_ratio : bool
            If *True*, the sampler will automatically terminate when the acceptance ratio falls behind an input
            threshold value.
        acceptance_ratio_threshold : float
            The acceptance ratio threshold below which sampling terminates if *terminate_at_acceptance_ratio* is *True*.
        """

        if paths is None:
            paths = Paths()

        super().__init__(
            paths=paths,
            prior_passer=prior_passer,
            initializer=InitializerPrior(),
            iterations_per_update=iterations_per_update,
        )

        self.terminate_at_acceptance_ratio = (
            self._config("settings", "terminate_at_acceptance_ratio", bool)
            if terminate_at_acceptance_ratio is None
            else terminate_at_acceptance_ratio
        )

        self.acceptance_ratio_threshold = (
            self._config("settings", "acceptance_ratio_threshold", float)
            if acceptance_ratio_threshold is None
            else acceptance_ratio_threshold
        )

        self.stagger_resampling_likelihood = (
            self._config("settings", "stagger_resampling_likelihood", bool)
            if stagger_resampling_likelihood is None
            else stagger_resampling_likelihood
        )

    class Fitness(NonLinearSearch.Fitness):
        def __init__(
            self,
            paths,
            analysis,
            model,
            samples_from_model,
            stagger_resampling_likelihood,
            terminate_at_acceptance_ratio,
            acceptance_ratio_threshold,
        ):

            super().__init__(
                paths=paths,
                analysis=analysis,
                model=model,
                samples_from_model=samples_from_model,
            )

            self.stagger_resampling_likelihood = stagger_resampling_likelihood
            self.stagger_accepted_samples = 0
            self.resampling_figure_of_merit = -1.0e99

            self.terminate_at_acceptance_ratio = terminate_at_acceptance_ratio
            self.acceptance_ratio_threshold = acceptance_ratio_threshold

            self.should_check_terminate = IntervalCounter(1000)

        def __call__(self, parameters, *kwargs):

            self.check_terminate_sampling()

            try:
                return self.figure_of_merit_from_parameters(parameters=parameters)
            except exc.FitException:
                return self.stagger_resampling_figure_of_merit()

        def figure_of_merit_from_parameters(self, parameters):
            """The figure of merit is the value that the non-linear search uses to sample parameter space. All Nested
            samplers use the log likelihood.
            """
            try:
                return self.log_likelihood_from_parameters(parameters=parameters)
            except exc.FitException:
                raise exc.FitException

        def stagger_resampling_figure_of_merit(self):
            """By default, when a fit raises an exception a log likelihood of -np.inf is returned, which leads the
            sampler to discard the sample.

            However, we found that this causes memory issues when running PyMultiNest. Therefore, we 'hack' a solution
            by not returning -np.inf (which leads the sample to be discarded) but instead a large negative float which
            is treated as a real sample (and does not lead too memory issues). The value returned is staggered to avoid
            all initial samples returning the same log likelihood and the non-linear search terminating."""

            if not self.stagger_resampling_likelihood:

                return self.resample_figure_of_merit

            else:

                if self.stagger_accepted_samples < 10:

                    self.stagger_accepted_samples += 1
                    self.resampling_figure_of_merit += 1e90

                    return self.resampling_figure_of_merit

                else:

                    return -1.0 * np.abs(self.resampling_figure_of_merit) * 10.0

        def check_terminate_sampling(self):
            """Automatically terminate nested sampling when the sampler's acceptance ratio falls below a specified
            value. This termimation is performed by returning all log likelihoods as the currently value of the maximum
            log likelihood sample. This will lead to unreliable probability density functions and error estimates.

            The reason to use this function is for stochastic likelihood functions the sampler can determine the
            highest log likelihood models in parameter space but get 'stuck', unable to terminate as it cannot get all
            live points to within a small likelihood range of one another. Without this feature on the sampler will not
            end and suffer an extremely low acceptance rate.

            This check is performed every 1000 samples."""

            if self.terminate_at_acceptance_ratio:

                try:
                    samples = self.samples_from_model(model=self.model)
                except Exception:
                    samples = None

                try:

                    if (
                        samples.acceptance_ratio < self.acceptance_ratio_threshold
                    ) or self.terminate_has_begun:

                        self.terminate_has_begun = True

                        return self.max_log_likelihood

                except ValueError:

                    pass

    @property
    def config_type(self):
        return conf.instance.nest

    def copy_with_name_extension(self, extension, remove_phase_tag=False):
        copy = super().copy_with_name_extension(
            extension=extension, remove_phase_tag=remove_phase_tag
        )
        copy.prior_passer = self.prior_passer
        copy.terminate_at_acceptance_ratio = self.terminate_at_acceptance_ratio
        copy.acceptance_ratio_threshold = self.acceptance_ratio_threshold
        copy.stagger_resampling_likelihood = self.stagger_resampling_likelihood
        return copy

    def fitness_function_from_model_and_analysis(self, model, analysis):

        return self.__class__.Fitness(
            paths=self.paths,
            model=model,
            analysis=analysis,
            samples_from_model=self.samples_via_sampler_from_model,
            stagger_resampling_likelihood=self.stagger_resampling_likelihood,
            terminate_at_acceptance_ratio=self.terminate_at_acceptance_ratio,
            acceptance_ratio_threshold=self.acceptance_ratio_threshold,
        )

    def samples_via_csv_json_from_model(self, model):

        parameters, log_likelihoods, log_priors, log_posteriors, weights = samp.load_from_table(
            filename=f"{self.paths.samples_path}/samples.csv", model=model
        )

        with open(f"{self.paths.samples_path}/info.json") as infile:
            samples_info = json.load(infile)

        return samp.NestSamples(
            model=model,
            parameters=parameters,
            log_likelihoods=log_likelihoods,
            log_priors=log_priors,
            weights=weights,
            log_evidence=samples_info["log_evidence"],
            total_samples=samples_info["total_samples"],
            unconverged_sample_size=samples_info["unconverged_sample_size"],
            number_live_points=samples_info["number_live_points"],
            time=samples_info["time"],
        )



from autofit.message_passing.factor_graphs.abstract import AbstractNode
from autofit.message_passing.factor_graphs.abstract import AbstractNode
from autofit.message_passing.factor_graphs.factor import FactorNode, FactorValue, Factor
from autofit.message_passing.factor_graphs.graph import DeterministicFactorNode, FactorGraph
from autofit.message_passing.factor_graphs.numerical import numerical_jacobian
from autofit.message_passing.factor_graphs.variable import Plate, Variable



import autofit as af

"""
This module create tags for phases settings that customize the analysis. We tag phases for two reasons:

    1) Tags describes the phase settings, making it explicit what analysis was used to create the results.

    2) Tags create unique output paths, ensuring that if you run multiple phases on the same data with different settings
       each non-linear search (e.g. Emcee) won't inadvertently use results generated via a different analysis method.
"""


class SettingsPhase(af.AbstractSettingsPhase):
    def __init__(self, data_trim_left=None, data_trim_right=None):

        super().__init__()

        self.data_trim_left = data_trim_left
        self.data_trim_right = data_trim_right

    @property
    def tag(self):

        """You may well have many more tags which appear here."""

        return (
            "settings"  # For every tag you add, you'll add it to this return statement
            + self.data_trim_left_tag
            + self.data_trim_right_tag
            # e.g. + your_own_tag
        )

    """
    This function generates a string we'll use to 'tag' a phase which uses this setting, thus ensuring results are
    output to a unique path.
    """

    @property
    def data_trim_left_tag(self):
        """Generate a data trim left tag, to customize phase names based on how much of the dataset is trimmed to 
        its left.
    
        This changes the phase name 'settings' as follows:
    
        data_trim_left = None -> settings
        data_trim_left = 2 -> settings__trim_left_2
        data_trim_left = 10 -> settings__trim_left_10
        """
        if self.data_trim_left is None:
            return ""
        return "__trim_left_" + str(self.data_trim_left)

    @property
    def data_trim_right_tag(self):
        """Generate a data trim right tag, to customize phase names based on how much of the dataset is trimmed to its right.
    
        This changes the phase name 'settings' as follows:
    
        data_trim_right = None -> settings
        data_trim_right = 2 -> settings__trim_right_2
        data_trim_right = 10 -> settings__trim_right_10
        """
        if self.data_trim_right is None:
            return ""
        return "__trim_right_" + str(self.data_trim_right)



import numpy as np

# The 'profiles.py' module is unchanged from the previous tutorial.


class Profile:
    def __init__(self, centre=0.0, intensity=0.01):
        """Represents an Abstract 1D profile.

        Parameters
        ----------
        centre : float
            The x coordinate of the profile centre.
        intensity : float
            Overall intensity normalisation of the profile.
        """

        self.centre = centre
        self.intensity = intensity


class Gaussian(Profile):
    def __init__(
        self,
        centre=0.0,  # <- PyAutoFit recognises these constructor arguments
        intensity=0.1,  # <- are the Gaussian's model parameters.
        sigma=0.01,
    ):
        """Represents a 1D Gaussian profile, which may be treated as a model-component of PyAutoFit the
        parameters of which are fitted for by a non-linear search.

        Parameters
        ----------
        centre : float
            The x coordinate of the profile centre.
        intensity : float
            Overall intensity normalisation of the Gaussian profile.
        sigma : float
            The sigma value controlling the size of the Gaussian.
        """

        super(Gaussian, self).__init__(centre=centre, intensity=intensity)

        self.sigma = sigma

    def profile_from_xvalues(self, xvalues):
        """
        Calculate the intensity of the profile on a line of Cartesian x coordinates.

        The input xvalues are translated to a coordinate system centred on the Gaussian, using its centre.

        Parameters
        ----------
        values : ndarray
            The x coordinates in the original reference frame of the grid.
        """
        transformed_xvalues = np.subtract(xvalues, self.centre)
        return np.multiply(
            np.divide(self.intensity, self.sigma * np.sqrt(2.0 * np.pi)),
            np.exp(-0.5 * np.square(np.divide(transformed_xvalues, self.sigma))),
        )


class Exponential(Profile):
    def __init__(
        self,
        centre=0.0,  # <- PyAutoFit recognises these constructor arguments are the model
        intensity=0.1,  # <- parameters of the Gaussian.
        rate=0.01,
    ):
        """Represents a 1D Exponential profile, which may be treated as a model-component of PyAutoFit the
        parameters of which are fitted for by a non-linear search.

        Parameters
        ----------
        centre : float
            The x coordinate of the profile centre.
        intensity : float
            Overall intensity normalisation of the Gaussian profile.
        ratw : float
            The decay rate controlling has fast the Exponential declines.
        """

        super(Exponential, self).__init__(centre=centre, intensity=intensity)

        self.rate = rate

    def profile_from_xvalues(self, xvalues):
        """
        Calculate the intensity of the profile on a line of Cartesian x coordinates.

        The input xvalues are translated to a coordinate system centred on the Exponential, using its centre.

        Parameters
        ----------
        values : ndarray
            The x coordinates in the original reference frame of the grid.
        """
        transformed_xvalues = np.subtract(xvalues, self.centre)
        return self.intensity * np.multiply(
            self.rate, np.exp(-1.0 * self.rate * abs(transformed_xvalues))
        )



import copy
import inspect
from functools import wraps
from numbers import Number
from random import random
from typing import Tuple, Optional

import numpy as np

from autofit import exc
from autofit.mapper import model
from autofit.mapper import model_mapper
from autofit.mapper.model import AbstractModel
from autofit.mapper.prior.deferred import DeferredArgument
from autofit.mapper.prior.prior import GaussianPrior
from autofit.mapper.prior.prior import TuplePrior, Prior, WidthModifier, Limits
from autofit.mapper.prior_model import collection
from autofit.mapper.prior_model import dimension_type as dim
from autofit.mapper.prior_model.attribute_pair import DeferredNameValue
from autofit.mapper.prior_model.attribute_pair import cast_collection, PriorNameValue, InstanceNameValue
from autofit.mapper.prior_model.recursion import DynamicRecursionCache
from autofit.mapper.prior_model.util import PriorModelNameValue
from autofit.text.formatter import TextFormatter


def check_assertions(func):
    @wraps(func)
    def wrapper(s, arguments):
        # noinspection PyProtectedMember
        failed_assertions = [
            assertion
            for assertion
            in s._assertions
            if assertion is False or assertion is not True and not assertion.instance_for_arguments(
                arguments
            )
        ]
        number_of_failed_assertions = len(failed_assertions)
        if number_of_failed_assertions > 0:
            name_string = "\n".join([
                assertion.name
                for assertion
                in failed_assertions
                if hasattr(assertion, "name") and assertion.name is not None
            ])
            raise exc.FitException(
                f"{number_of_failed_assertions} assertions failed!\n{name_string}"
            )

        return func(s, arguments)

    return wrapper


class AbstractPriorModel(AbstractModel):
    """
    Abstract model that maps a set of priors to a particular class. Must be
    overridden by any prior model so that the model mapper recognises its prior \
    model attributes.
    @DynamicAttrs
    """

    def __init__(self):
        super().__init__()
        self._assertions = list()

    def add_assertion(self, assertion, name=None):
        """
        Assert that some relationship holds between physical values associated with
        priors at the point an instance is created. If this fails a FitException is
        raised causing the model to be re-sampled.
        Parameters
        ----------
        assertion
            An assertion that one prior must be greater than another.
        name
            A name describing the assertion that is logged when it is violated.
        """
        if assertion is True:
            return
        try:
            assertion.name = name
        except AttributeError:
            pass
        self._assertions.append(assertion)

    @property
    def name(self):
        return self.__class__.__name__

    # noinspection PyUnusedLocal
    @staticmethod
    def from_object(t, *args, **kwargs):
        if inspect.isclass(t):
            from .prior_model import PriorModel

            obj = object.__new__(PriorModel)
            obj.__init__(t, **kwargs)
        elif isinstance(t, list) or isinstance(t, dict):
            obj = object.__new__(
                collection.CollectionPriorModel
            )
            obj.__init__(t)
        else:
            obj = t
        return obj

    def instance_from_unit_vector(self, unit_vector, assert_priors_in_limits=True):
        """
        Creates a ModelInstance, which has an attribute and class instance corresponding
        to every PriorModel attributed to this instance.
        This method takes as input a unit vector of parameter values, converting each to
        physical values via their priors.
        Parameters
        ----------
        unit_vector: [float]
            A unit hypercube vector that is mapped to an instance of physical values via the priors.
        Returns
        -------
        model_instance : autofit.mapper.model.ModelInstance
            An object containing reconstructed model_mapper instances
        Raises
        ------
        exc.FitException
            If any assertion attached to this object returns False.
        """
        arguments = dict(
            map(
                lambda prior_tuple, unit: (
                    prior_tuple.prior,
                    prior_tuple.prior.value_for(unit),
                ),
                self.prior_tuples_ordered_by_id,
                unit_vector,
            )
        )

        return self.instance_for_arguments(arguments, assert_priors_in_limits=assert_priors_in_limits)

    @property
    @cast_collection(PriorNameValue)
    def unique_prior_tuples(self):
        """
        Returns
        -------
        prior_tuple_dict: [(Prior, PriorTuple)]
            The set of all priors associated with this mapper
        """
        return {
            prior_tuple[1]: prior_tuple
            for prior_tuple in self.attribute_tuples_with_type(Prior)
        }.values()

    @property
    def unique_promise_tuples(self):
        from autofit.mapper.prior.promise import AbstractPromise

        return {
            prior_tuple[1]: prior_tuple
            for prior_tuple in self.attribute_tuples_with_type(AbstractPromise)
        }.values()

    @property
    @cast_collection(PriorNameValue)
    def prior_tuples_ordered_by_id(self):
        """
        Returns
        -------
        priors: [Prior]
            An ordered list of unique priors associated with this mapper
        """
        return sorted(
            list(self.unique_prior_tuples), key=lambda prior_tuple: prior_tuple.prior.id
        )

    def vector_from_unit_vector(self, unit_vector):
        """
        Parameters
        ----------
        unit_vector: [float]
            A unit hypercube vector
        Returns
        -------
        values: [float]
            A vector with values output by priors
        """
        return list(
            map(
                lambda prior_tuple, unit: prior_tuple.prior.value_for(unit),
                self.prior_tuples_ordered_by_id,
                unit_vector,
            )
        )

    def random_unit_vector_within_limits(self, lower_limit=0.0, upper_limit=1.0):
        """ Generate a random vector of unit values by drawing uniform random values between 0 and 1.
        Returns
        -------
        unit_values: [float]
            A list of unit values constructed by taking random values from each prior.
        """
        return list(np.random.uniform(low=lower_limit, high=upper_limit, size=self.prior_count))

    def random_vector_from_priors_within_limits(self, lower_limit, upper_limit):
        """ Generate a random vector of physical values by drawing uniform random values between an input lower and
        upper limit and using the model priors to map them from unit values to physical values.
        This is used for MCMC initialization, whereby the starting points of a walker(s) is confined to a restricted
        range of prior space. In particular, it is used for generate the "ball" initialization of Emcee.
        Returns
        -------
        physical_values: [float]
            A list of physical values constructed by taking the mean possible value from
            each prior.
        """

        while True:

            vector = self.vector_from_unit_vector(
                list(np.random.uniform(low=lower_limit, high=upper_limit, size=self.prior_count))
            )

            try:
                self.instance_from_vector(vector=vector)
                return vector
            except exc.PriorLimitException:
                pass

    @property
    def random_vector_from_priors(self):
        """ Generate a random vector of physical values by drawing uniform random values between 0 and 1 and using
        the model priors to map them from unit values to physical values.
        Returns
        -------
        physical_values: [float]
            A list of physical values constructed by taking random values from each prior.
        """
        return self.random_vector_from_priors_within_limits(lower_limit=0.0, upper_limit=1.0)

    @property
    def physical_values_from_prior_medians(self):
        """
        Returns
        -------
        physical_values: [float]
            A list of physical values constructed by taking the mean possible value from
            each prior.
        """
        return self.vector_from_unit_vector([0.5] * len(self.unique_prior_tuples))

    def instance_from_vector(
            self,
            vector,
            assert_priors_in_limits=True
    ):
        """
        Creates a ModelInstance, which has an attribute and class instance corresponding
        to every PriorModel attributed to this instance.
        This method takes as input a physical vector of parameter values, thus omitting
        the use of priors.
        Parameters
        ----------
        vector: [float]
            A vector of physical parameter values that is mapped to an instance.
        assert_priors_in_limits
            If True it is checked that the physical values of priors are within set limits
        Returns
        -------
        model_instance : autofit.mapper.model.ModelInstance
            An object containing reconstructed model_mapper instances
        """
        arguments = dict(
            map(
                lambda prior_tuple, physical_unit: (prior_tuple.prior, physical_unit),
                self.prior_tuples_ordered_by_id,
                vector,
            )
        )

        return self.instance_for_arguments(
            arguments,
            assert_priors_in_limits=assert_priors_in_limits
        )

    def mapper_from_partial_prior_arguments(self, arguments):
        """
        Creates a new model mapper from a dictionary mapping_matrix existing priors to
        new priors, keeping existing priors where no mapping is provided.
        Parameters
        ----------
        arguments: {Prior: Prior}
            A dictionary mapping_matrix priors to priors
        Returns
        -------
        model_mapper: ModelMapper
            A new model mapper with updated priors.
        """
        original_prior_dict = {prior: prior for prior in self.priors}
        return self.mapper_from_prior_arguments({**original_prior_dict, **arguments})

    def mapper_from_prior_arguments(self, arguments):
        """
        Creates a new model mapper from a dictionary mapping_matrix existing priors to
        new priors.
        Parameters
        ----------
        arguments: {Prior: Prior}
            A dictionary mapping_matrix priors to priors
        Returns
        -------
        model_mapper: ModelMapper
            A new model mapper with updated priors.
        """
        mapper = copy.deepcopy(self)

        for prior_model_tuple in self.prior_model_tuples:
            setattr(
                mapper,
                prior_model_tuple.name,
                prior_model_tuple.prior_model.gaussian_prior_model_for_arguments(
                    arguments
                ),
            )

        return mapper

    def mapper_from_gaussian_tuples(
            self,
            tuples,
            a=None,
            r=None,
            use_errors=True,
            use_widths=True,
            no_limits=False
    ):
        """
        Creates a new model mapper from a list of floats describing the mean values
        of gaussian priors. The widths of the new priors are taken from the
        width_config. The new gaussian priors must be provided in the same order as
        the priors associated with model.
        If a is not None then all priors are created with an absolute width of a.
        If r is not None then all priors are created with a relative width of r.
        Parameters
        ----------
        no_limits
            If True generated priors have infinite limits
        r
            The relative width to be assigned to gaussian priors
        a
            print(tuples[i][1], width)
            The absolute width to be assigned to gaussian priors
        use_errors : bool
            If True, the passed errors of the model components estimated in a previous non-linear search (computed
            at the prior_passer.sigma value) are used to set the pass Gaussian Prior sigma value (if both width and
            passed errors are used, the maximum of these two values are used).
        use_widths : bool
            If True, the minimum prior widths specified in the json_prior configs of the model components are used to
            set the passed Gaussian Prior sigma value (if both widths and passed errors are used, the maximum of
            these two values are used).
        tuples
            A list of tuples each containing the mean and width of a prior
        Returns
        -------
        mapper: ModelMapper
            A new model mapper with all priors replaced by gaussian priors.
        """

        prior_tuples = self.prior_tuples_ordered_by_id
        prior_class_dict = self.prior_class_dict
        arguments = {}

        for i, prior_tuple in enumerate(prior_tuples):
            prior = prior_tuple.prior
            cls = prior_class_dict[prior]
            mean, sigma = tuples[i]

            name = prior_tuple.name
            # Use the name of the collection for configuration when a prior's name
            # is just a number (i.e. its position in a collection)
            if name.isdigit():
                name = self.path_for_prior(prior_tuple.prior)[-2]

            width_modifier = WidthModifier.for_class_and_attribute_name(cls, name)

            if a is not None and r is not None:
                raise exc.PriorException(
                    "Width of new priors cannot be both relative and absolute."
                )
            if a is not None:
                width = a
            elif r is not None:
                width = r * mean
            else:
                width = width_modifier(mean)

            if no_limits:
                limits = (float("-inf"), float("inf"))
            else:
                try:
                    limits = Limits.for_class_and_attributes_name(
                        cls,
                        name
                    )
                except exc.PriorException:
                    limits = prior.limits

            if use_errors and not use_widths:
                sigma = tuples[i][1]
            elif not use_errors and use_widths:
                sigma = width
            elif use_errors and use_widths:
                sigma = max(tuples[i][1], width)
            else:
                raise exc.PriorException("use_passed_errors and use_widths are both False, meeaning there is no "
                                         "way to pass priors to set up the new model's Gaussian Priors.")

            arguments[prior] = GaussianPrior(
                mean,
                sigma,
                *limits
            )

        return self.mapper_from_prior_arguments(arguments)

    def instance_from_prior_medians(self):
        """
        Creates a list of physical values from the median values of the priors.
        Returns
        -------
        physical_values : [float]
            A list of physical values
        """
        return self.instance_from_unit_vector(
            unit_vector=[0.5] * len(self.prior_tuples)
        )

    def log_priors_from_vector(
            self,
            vector: [float],
    ):
        """
        Compute the log priors of every parameter in a vector, using the Prior of every parameter.
        The log prior values are used by Emcee to map the log likelihood to the poserior of the model.
        Parameters
        ----------
        vector : [float]
            A vector of physical parameter values.
        Returns
        -------
        log_priors : []
            An list of the log prior value of every parameter.
        """
        return list(
            map(
                lambda prior_tuple, value: prior_tuple.prior.log_prior_from_value(value=value),
                self.prior_tuples_ordered_by_id,
                vector,
            )
        )

    def random_instance(self):
        """
        Creates a random instance of the model.
        """
        return self.instance_from_unit_vector(
            unit_vector=[random() for _ in self.prior_tuples]
        )

    @staticmethod
    @DynamicRecursionCache()
    def from_instance(instance, model_classes=tuple()):
        """
        Recursively create an prior object model from an object model.
        Parameters
        ----------
        model_classes
        instance
            A dictionary, list, class instance or model instance
        Returns
        -------
        abstract_prior_model
            A concrete child of an abstract prior model
        """

        if isinstance(instance, list):
            result = collection.CollectionPriorModel(
                [
                    AbstractPriorModel.from_instance(item, model_classes=model_classes)
                    for item in instance
                ]
            )
        elif isinstance(instance, model.ModelInstance):
            result = model_mapper.ModelMapper()
            for key, value in instance.dict.items():
                setattr(
                    result,
                    key,
                    AbstractPriorModel.from_instance(
                        value, model_classes=model_classes
                    ),
                )
        elif isinstance(instance, dict):
            result = collection.CollectionPriorModel(
                {
                    key: AbstractPriorModel.from_instance(
                        value, model_classes=model_classes
                    )
                    for key, value in instance.items()
                }
            )
        elif isinstance(instance, (dim.DimensionType, np.ndarray)):
            return instance
        else:
            from .prior_model import PriorModel

            try:
                result = PriorModel(
                    instance.__class__,
                    **{
                        key: AbstractPriorModel.from_instance(
                            value, model_classes=model_classes
                        )
                        for key, value in instance.__dict__.items()
                        if key != "cls"
                    },
                )
            except AttributeError:
                return instance
        if any([isinstance(instance, cls) for cls in model_classes]):
            return result.as_model()
        return result

    @property
    @cast_collection(PriorNameValue)
    def direct_prior_tuples(self):
        return self.direct_tuples_with_type(Prior)

    @property
    @cast_collection(InstanceNameValue)
    def direct_instance_tuples(self):
        return self.direct_tuples_with_type(float)

    @property
    @cast_collection(PriorModelNameValue)
    def prior_model_tuples(self):
        return self.direct_tuples_with_type(AbstractPriorModel)

    @property
    @cast_collection(PriorNameValue)
    def tuple_prior_tuples(self):
        """
        Returns
        -------
        tuple_prior_tuples: [(String, TuplePrior)]
        """
        return self.direct_tuples_with_type(TuplePrior)

    @property
    @cast_collection(PriorNameValue)
    def direct_prior_tuples(self):
        """
        Returns
        -------
        direct_priors: [(String, Prior)]
        """
        return self.direct_tuples_with_type(Prior)

    @property
    @cast_collection(DeferredNameValue)
    def direct_deferred_tuples(self):
        return self.direct_tuples_with_type(DeferredArgument)

    @property
    @cast_collection(PriorNameValue)
    def prior_tuples(self):
        """
        Returns
        -------
        priors: [(String, Prior))]
        """
        # noinspection PyUnresolvedReferences
        return self.attribute_tuples_with_type(Prior)

    @property
    @cast_collection(PriorModelNameValue)
    def direct_prior_model_tuples(self):
        return self.direct_tuples_with_type(AbstractPriorModel)

    def __eq__(self, other):
        return (
                isinstance(other, AbstractPriorModel)
                and self.direct_prior_model_tuples == other.direct_prior_model_tuples
        )

    @property
    @cast_collection(InstanceNameValue)
    def instance_tuples(self):
        """
        Returns
        -------
        instances: [(String, instance)]
        """
        return self.attribute_tuples_with_type(float, ignore_class=Prior)

    @property
    def prior_class_dict(self):
        from autofit.mapper.prior_model.annotation import AnnotationPriorModel

        d = {prior[1]: self.cls for prior in self.prior_tuples}
        for prior_model in self.prior_model_tuples:
            if not isinstance(prior_model[1], AnnotationPriorModel):
                d.update(prior_model[1].prior_class_dict)
        return d

    def _instance_for_arguments(self, arguments):
        raise NotImplementedError()

    def instance_for_arguments(
            self,
            arguments,
            assert_priors_in_limits=True
    ):
        """
        Create an instance of the model for a set of arguments
        Parameters
        ----------
        assert_priors_in_limits
            If true it is asserted that the physical values that replace piors are
            within their limits
        arguments: {Prior: float}
            Dictionary mapping_matrix priors to attribute analysis_path and value pairs
        Returns
        -------
            An instance of the class
        """
        if self.promise_count > 0:
            raise exc.PriorException(
                "All promises must be populated prior to instantiation"
            )
        if assert_priors_in_limits:
            for prior, value in arguments.items():
                if isinstance(value, Number):
                    prior.assert_within_limits(value)
        return self._instance_for_arguments(
            arguments
        )

    @property
    def prior_count(self):
        return len(self.unique_prior_tuples)

    @property
    def promise_count(self):
        return len(self.unique_promise_tuples)

    @property
    def variable_promise_count(self):
        return len([
            value for key, value in
            self.unique_promise_tuples
            if not value.is_instance
        ])

    @property
    def priors(self):
        return [prior_tuple.prior for prior_tuple in self.prior_tuples]

    def name_for_prior(self, prior):
        for prior_model_name, prior_model in self.direct_prior_model_tuples:
            prior_name = prior_model.name_for_prior(prior)
            if prior_name is not None:
                return "{}_{}".format(prior_model_name, prior_name)
        prior_tuples = self.prior_tuples
        for name, p in prior_tuples:
            if p == prior:
                return name

    def __hash__(self):
        return self.id

    def __add__(self, other):
        result = copy.deepcopy(self)

        for key, value in other.__dict__.items():
            if not hasattr(result, key) or isinstance(value, Prior):
                setattr(result, key, value)
                continue
            result_value = getattr(result, key)
            if isinstance(value, AbstractPriorModel):
                if isinstance(result_value, AbstractPriorModel):
                    setattr(result, key, result_value + value)
                else:
                    setattr(result, key, value)

        return result

    def copy_with_fixed_priors(self, instance, excluded_classes=tuple()):
        """
        Recursively overwrite priors in the mapper with instance values from the
        instance except where the containing class is the descendant of a listed class.
        Parameters
        ----------
        excluded_classes
            Classes that should be left model
        instance
            The best fit from the previous phase
        """
        mapper = copy.deepcopy(self)
        transfer_classes(instance, mapper, excluded_classes)
        return mapper

    @property
    def path_priors_tuples(self):
        path_priors_tuples = self.path_instance_tuples_for_class(Prior)
        return sorted(path_priors_tuples, key=lambda item: item[1].id)

    def path_for_prior(self, prior: Prior) -> Optional[Tuple[str]]:
        """
        Find a path that points at the given tuple.
        Returns the first path or None if no path is found.
        Parameters
        ----------
        prior
            A prior representing what's known about some dimension of the model.
        Returns
        -------
        A path, a series of attributes that point to one location of the prior.
        """
        for path, path_prior in self.path_priors_tuples:
            if path_prior == prior:
                return path
        return None

    @property
    def path_float_tuples(self):
        return self.path_instance_tuples_for_class(float, ignore_class=Prior)

    @property
    def unique_prior_paths(self):
        unique = {item[1]: item for item in self.path_priors_tuples}.values()
        return [item[0] for item in sorted(unique, key=lambda item: item[1].id)]

    @property
    def prior_prior_model_dict(self):
        """
        Returns
        -------
        prior_prior_model_dict: {Prior: PriorModel}
            A dictionary mapping priors to associated prior models. Each prior will only
            have one prior model; if a prior is shared by two prior models then one of
            those prior models will be in this dictionary.
        """
        return {
            prior: prior_model[1]
            for prior_model in self.prior_model_tuples + [("model", self)]
            for _, prior in prior_model[1].prior_tuples
        }

    @property
    def info(self) -> str:
        """
        Use the priors that make up the model_mapper to generate information on each
        parameter of the overall model.
        This information is extracted from each priors *model_info* property.
        """
        from autofit.mapper.prior import AbstractPromise
        formatter = TextFormatter()

        for t in self.path_instance_tuples_for_class((
                Prior, float, AbstractPromise, tuple
        )):
            formatter.add(t)

        return formatter.text

    @property
    def parameter_names(self) -> [str]:
        """The param_names vector is a list each parameter's analysis_path, and is used
        for *GetDist* visualization.
        The parameter names are determined from the class instance names of the
        model_mapper. Latex tags are properties of each model class."""
        return [
            self.name_for_prior(
                prior
            )
            for _, prior
            in self.prior_tuples_ordered_by_id
        ]


def transfer_classes(instance, mapper, model_classes=None):
    """
    Recursively overwrite priors in the mapper with instance values from the
    instance except where the containing class is the descendant of a listed class.
    Parameters
    ----------
    model_classes
        Classes whose descendants should not be overwritten
    instance
        The best fit from the previous phase
    mapper
        The prior model from the previous phase
    """
    from autofit.mapper.prior_model.annotation import AnnotationPriorModel

    model_classes = model_classes or []
    for key, instance_value in instance.__dict__.items():
        try:
            mapper_value = getattr(mapper, key)
            if isinstance(mapper_value, Prior) or isinstance(
                    mapper_value, AnnotationPriorModel
            ):
                setattr(mapper, key, instance_value)
                continue
            if not any(isinstance(instance_value, cls) for cls in model_classes):
                try:
                    transfer_classes(instance_value, mapper_value, model_classes)
                except AttributeError:
                    setattr(mapper, key, instance_value)
        except AttributeError:
            pass



from abc import ABC, abstractmethod
from collections import ChainMap
from typing import Union, Dict, List, Tuple, cast, Collection

import numpy as np

from autofit.message_passing.factor_graphs.variable import Variable, Plate


class AbstractNode(ABC):
    _deterministic_variables: Union[
        Dict[str, Variable],
        ChainMap
    ]

    def __init__(
            self,
            *args: Variable,
            **kwargs: Variable
    ):
        """
        A node in a factor graph

        Parameters
        ----------
        args
            Positional arguments passed to the factor
        kwargs
            Key word arguments passed to the value
        """
        self._variables = {
            v.name: v
            for v
            in args + tuple(
                kwargs.values()
            )
        }

        self._args = args
        self._kwargs = kwargs

    @property
    @abstractmethod
    def deterministic_variables(self):
        return self._deterministic_variables

    @property
    @abstractmethod
    def name(self) -> str:
        """
        A name for this object
        """

    @property
    def call_signature(self) -> str:
        """
        The apparent signature of this object
        """
        args = ", ".join(self.arg_names)
        kws = ", ".join(map("{0[0]}={0[1]}".format, self.kwarg_names))
        call_strs = []
        if args:
            call_strs.append(args)
        if kws:
            call_strs.extend(['*', kws])
        call_str = ", ".join(call_strs)
        call_sig = f"{self.name}({call_str})"
        return call_sig

    @property
    def arg_names(self) -> List[str]:
        """
        The names of the variables passed as positional arguments
        """
        return [
            arg.name
            for arg
            in self._args
        ]

    @property
    def kwarg_names(self) -> List[str]:
        """
        The names of the variables passed as keyword arguments
        """
        return [
            arg.name
            for arg
            in self._kwargs.values()
        ]

    @property
    def all_variables(self) -> Dict[str, Variable]:
        """
        A dictionary of variables associated with this node
        """
        return {
            **self._variables,
            **self._deterministic_variables
        }

    def _broadcast(
            self,
            plate_inds: np.ndarray,
            value: np.ndarray
    ) -> np.ndarray:
        """
        Ensure the shape of the data matches the shape of the plates

        Parameters
        ----------
        plate_inds
            The indices of the plates of some factor within this node
        value
            Some data

        Returns
        -------
        The data reshaped
        """
        shape = np.shape(value)
        shift = len(shape) - plate_inds.size

        assert shift in {0, 1}
        newshape = np.ones(self.ndim + shift, dtype=int)
        newshape[:shift] = shape[:shift]
        newshape[shift + plate_inds] = shape[shift:]

        return np.reshape(value, newshape)

    @property
    def plates(self) -> Tuple[Plate]:
        """
        A tuple of the set of all plates in this graph, ordered by id
        """
        return tuple(sorted(set(
            cast(Plate, plate)
            for variable
            in self.all_variables.values()
            for plate in variable.plates
        )))

    @property
    def ndim(self) -> int:
        """
        The number of plates contained within this graph's variables

        That is, the total dimensions of those variables.
        """
        return len(self.plates)

    def _match_plates(
            self,
            plates: Collection[Plate]
    ) -> np.ndarray:
        """
        Find indices plates from some factor in the collection of
        plates associated with this node.

        Parameters
        ----------
        plates
            Plates from some other node

        Returns
        -------
        An array of plate indices
        """
        return np.array([self.plates.index(p) for p in plates], dtype=int)



import autofit as af

def test__phase_settings_tag():

    settings = af.AbstractSettingsPhase(log_likelihood_cap=None)
    assert settings.log_likelihood_cap_tag == ""

    settings = af.AbstractSettingsPhase(log_likelihood_cap=10.0)
    assert settings.log_likelihood_cap_tag == "__lh_cap_10.0"

    settings = af.AbstractSettingsPhase(log_likelihood_cap=200.0001)
    assert settings.log_likelihood_cap_tag == "__lh_cap_200.0"


import numpy as np

# TODO: Use autofit class?
from scipy import stats


def _gaussian(x, centre, intensity, sigma):
    return Gaussian(centre=centre, intensity=intensity, sigma=sigma)(x)


# TODO: use autofit likelihood
def _likelihood(z, y):
    _norm = stats.norm(loc=0, scale=1.)
    return _norm.logpdf(z - y)


class Profile:
    def __init__(self, centre=0.0, intensity=0.01):
        """Represents an Abstract 1D profile.

        Parameters
        ----------
        centre : float
            The x coordinate of the profile centre.
        intensity : float
            Overall intensity normalisation of the profile.
        """
        self.centre = centre
        self.intensity = intensity


class Gaussian(Profile):
    def __init__(
            self,
            centre=0.0,  # <- PyAutoFit recognises these constructor arguments
            intensity=0.1,  # <- are the Gaussian's model parameters.
            sigma=0.01,
    ):
        """Represents a 1D Gaussian profile, which may be treated as a model-component of PyAutoFit the
        parameters of which are fitted for by a non-linear search.

        Parameters
        ----------
        centre : float
            The x coordinate of the profile centre.
        intensity : float
            Overall intensity normalisation of the Gaussian profile.
        sigma : float
            The sigma value controlling the size of the Gaussian.
        """
        super().__init__(centre=centre, intensity=intensity)
        self.sigma = sigma  # We still need to set sigma for the Gaussian, of course.

    def __call__(self, xvalues):
        """
        Calculate the intensity of the profile on a line of Cartesian x coordinates.

        The input xvalues are translated to a coordinate system centred on the Gaussian, using its centre.

        Parameters
        ----------
        xvalues : ndarray
            The x coordinates in the original reference frame of the grid.
        """
        transformed_xvalues = np.subtract(xvalues, self.centre)
        return np.multiply(
            np.divide(self.intensity, self.sigma * np.sqrt(2.0 * np.pi)),
            np.exp(-0.5 * np.square(np.divide(transformed_xvalues, self.sigma))),
        )


def make_data(
        gaussian,
        n_observations
):
    x = np.arange(n_observations)
    model_line = gaussian(xvalues=x)
    signal_to_noise_ratio = 25.0
    noise = np.random.normal(0.0, 1.0 / signal_to_noise_ratio, n_observations)
    y = model_line + noise
    return x, y



import autofit as af
from howtofit.chapter_1_introduction.tutorial_7_phase_customization.src.dataset.dataset import (
    Dataset,
)
from howtofit.chapter_1_introduction.tutorial_7_phase_customization.src.phase.result import (
    Result,
)
from howtofit.chapter_1_introduction.tutorial_7_phase_customization.src.phase.analysis import (
    Analysis,
)
from howtofit.chapter_1_introduction.tutorial_7_phase_customization.src.phase.meta_dataset import (
    MetaDataset,
)
from howtofit.chapter_1_introduction.tutorial_7_phase_customization.src.phase.settings import (
    SettingsPhase,
)


"""
The phase module has new features not included in tutorial 6, which customize the dataset that is fitted and tag
the output path of the results.
"""


class Phase(af.AbstractPhase):

    """
    Because we now have multiple profiles in our model, we have renamed 'gaussian' to 'profiles'. As before,
    PyAutoFit uses this information to map the input Profile classes to a model instance when performing a fit.
    """

    profiles = af.PhaseProperty("profiles")

    Result = Result

    @af.convert_paths
    def __init__(self, paths, profiles, settings, search):
        """
        A phase which fits a model composed of multiple profiles (Gaussian, Exponential) using a non-linear search.

        Parameters
        ----------
        paths : af.Paths
            Handles the output directory structure.
        profiles : [profiles.Profile]
            The model components (e.g. Gaussian, Exponenial) fitted by this phase.
        search: class
            The class of a non_linear search
        settings : SettingsPhase
            The collection of settings of the phase used to augment the data that is fitted and tag the output path.
        """

        """
        Here, we create a 'tag' for our phase. If we use an optional phase setting to alter the dataset we fit (here,
        a data_trim_ variable), we want to 'tag' the phase such that results are output to a unique
        directory whose names makes it explicit how the dataset was changed.

        If this setting is off, the tag is an empty string and thus the directory structure is not changed.
        """

        paths.tag = settings.tag  # The phase_tag must be manually added to the phase.

        super().__init__(paths=paths, search=search)

        self.profiles = profiles

        """
        Phase settings alter the dataset that is fitted, however a phase does not have access to the dataset until it
        is run (e.g. the run method below is passed the dataset). In order for a phase to use its input phase
        settings to create the dataset it fits, these settings are stored in the 'meta_dataset' attribute and used
        when the 'run' and 'make_analysis' methods are called.
        """

        self.meta_dataset = MetaDataset(settings=settings)

    def run(self, dataset: Dataset, mask):
        """
        Pass a dataset to the phase, running the phase and non-linear search.

        Parameters
        ----------
        dataset: aa.Dataset
            The dataset fitted by the phase, as defined in the 'dataset.py' module.
        mask: Mask
            The mask used for the analysis.

        Returns
        -------
        result: AbstractPhase.Result
            A result object comprising information on the non-linear search and the maximum likelihood model.
        """

        analysis = self.make_analysis(dataset=dataset, mask=mask)

        result = self.run_analysis(analysis=analysis)

        return self.make_result(result=result, analysis=analysis)

    def make_analysis(self, dataset, mask):
        """
        Create an Analysis object, which creates the dataset and contains the functions which perform the fit.

        Parameters
        ----------
        dataset: aa.Dataset
            The dataset fitted by the phase, as defined in the 'dataset.py' module.

        Returns
        -------
        analysis : Analysis
            An analysis object that the non-linear search calls to determine the fit log_likelihood for a given model
            instance.
        """

        """
        Here, the meta_dataset is used to create the masked dataset that is fitted. If the data_trim_left and / or
        data_trim_right settings are passed into the phase, the function below uses them to alter the masked dataset.

        Checkout 'meta_dataset.py' for more details.
        """

        masked_dataset = self.meta_dataset.masked_dataset_from_dataset_and_mask(
            dataset=dataset, mask=mask
        )

        return Analysis(
            masked_dataset=masked_dataset, image_path=self.search.paths.image_path
        )

    def make_result(self, result, analysis):
        return self.Result(
            samples=result.samples,
            previous_model=self.model,
            search=self.search,
            analysis=analysis,
        )



from abc import ABC

from autofit.mapper.prior.compound import CompoundPrior
from autofit.mapper.prior_model.abstract import AbstractPriorModel


class ComparisonAssertion(CompoundPrior, ABC):
    def __init__(
            self,
            lower,
            greater,
            name=None
    ):
        super().__init__(lower, greater)
        self._name = name

    def __gt__(self, other):
        return CompoundAssertion(self, self.left > other)

    def __lt__(self, other):
        return CompoundAssertion(self, self.right < other)

    def __ge__(self, other):
        return CompoundAssertion(self, self.left >= other)

    def __le__(self, other):
        return CompoundAssertion(self, self.right <= other)


class GreaterThanLessThanAssertion(ComparisonAssertion):
    def _instance_for_arguments(self, arguments):
        """
        Assert that the value in the dictionary associated with the lower
        prior is lower than the value associated with the greater prior.

        Parameters
        ----------
        arguments
            A dictionary mapping priors to physical values.

        Raises
        ------
        FitException
            If the assertion is not met
        """
        lower = self.left_for_arguments(arguments)
        greater = self.right_for_arguments(arguments)
        return lower < greater


class GreaterThanLessThanEqualAssertion(ComparisonAssertion):
    def _instance_for_arguments(self, arguments):
        """
        Assert that the value in the dictionary associated with the lower
        prior is lower than the value associated with the greater prior.

        Parameters
        ----------
        arguments
            A dictionary mapping priors to physical values.

        Raises
        ------
        FitException
            If the assertion is not met
        """
        return self.left_for_arguments(
            arguments
        ) <= self.right_for_arguments(
            arguments
        )


class CompoundAssertion(AbstractPriorModel):
    def __init__(
            self,
            assertion_1,
            assertion_2,
            name=None
    ):
        super().__init__()
        self.assertion_1 = assertion_1
        self.assertion_2 = assertion_2
        self._name = name

    def _instance_for_arguments(self, arguments):
        return self.assertion_1.instance_for_arguments(
            arguments
        ) and self.assertion_2.instance_for_arguments(
            arguments
        )


def unwrap(obj):
    try:
        return obj._value
    except AttributeError:
        return obj



import itertools


class ModelObject:
    _ids = itertools.count()

    def __init__(self):
        self.id = next(self._ids)

    @property
    def component_number(self):
        return self.id

    def __hash__(self):
        return self.id

    def __eq__(self, other):
        try:
            return self.id == other.id
        except AttributeError:
            return False



# %%
"""
Tutorial 8: Aggregator Part 1
=============================

After fitting a large suite of data with the same pipeline, the aggregator allows us to load the results and
manipulate / plot them using a Python script or Jupyter notebook.

To begin, we need a set of results that we want to analyse using the aggregator. We'll create this by fitting 3
different data-sets. Each dataset is a single Gaussian and we'll fit them using a single Gaussian model.
"""

# %%
#%matplotlib inline

from autoconf import conf
import autofit as af
from howtofit.chapter_1_introduction.tutorial_8_aggregator import src as htf

import numpy as np
from pyprojroot import here

workspace_path = str(here())
print("Workspace Path: ", workspace_path)

# %%
"""
Setup the configs as we did in the previous tutorial, as well as the output folder for our non-linear search.
"""

# %%
conf.instance = conf.Config(
    config_path=f"{workspace_path}/howtofit/config",
    output_path=f"{workspace_path}/howtofit/output",
)

# %%
"""
Here, for each dataset we are going to set up the correct path, load it, create its mask and fit it with the phase above.

We want our results to be in a folder specific to the dataset. We'll use the dataset_name string to do this. Lets
create a list of all 3 of our dataset names.

We'll also pass these names to the dataset when we create it - the name will be used by the aggregator to name the file
the data is stored. More importantly, the name will be accessible to the aggregator, and we will use it to label 
figures we make via the aggregator.
"""

# %%
from howtofit.simulators.chapter_1 import gaussian_x1_0
from howtofit.simulators.chapter_1 import gaussian_x1_1
from howtofit.simulators.chapter_1 import gaussian_x1_2

dataset_names = ["gaussian_x1_0", "gaussian_x1_1", "gaussian_x1_2"]
datas = [gaussian_x1_0.data, gaussian_x1_1.data, gaussian_x1_2.data]
noise_maps = [gaussian_x1_0.noise_map, gaussian_x1_1.noise_map, gaussian_x1_2.noise_map]

# %%
"""
We can also attach information to the model-fit, by setting up an info dictionary. 

Information about our model-fit (e.g. the dataset) that isn't part of the model-fit is made accessible to the 
aggregator. For example, below we write info on the dataset's data of observation and exposure time.
"""

# %%
info = {"date_of_observation": "01-02-18", "exposure_time": 1000.0}

# %%
"""
This for loop runs over every dataset, checkout the comments below for how we set up the path structure.
"""

# %%
for index in range(len(datas)):

    """The code below creates the dataset and mask as per usual."""

    dataset = htf.Dataset(data=datas[index], noise_map=noise_maps[index])

    mask = np.full(fill_value=False, shape=dataset.data.shape)

    """
    Here, we create a phase as normal. However, we also include an input parameter 'folders'. The phase folders
    define the names of folders that the phase goes in. For example, if a phase goes to the path:

        '/path/to/autofit_workspace/output/phase_name/'

    A phase folder with the input 'phase_folder' edits this path to:

        '/path/to/autofit_workspace/output/phase_folder/phase_name/'

    You can input multiple phase folders, for example 'folders=['folder_0', 'folder_1'] would create the path:

        '/path/to/autofit_workspace/output/folder_0/folder_1/phase_name/'

    Below, we use the data_name, so our results go in a folder specific to the dataset, e.g:

        '/path/to/autofit_workspace/output/gaussian_x1_0/phase_t8/'
    """

    print(
        f"Emcee has begun running - checkout the "
        f"autofit_workspace/howtofit/chapter_1_introduction/output/{dataset_names[index]}/phase_t8 folder for live "
        f"output of the results. This Jupyter notebook cell with progress once Emcee has completed - this could take a "
        f"few minutes!"
    )

    phase = htf.Phase(
        phase_name="phase_t8",
        folders=[dataset_names[index]],
        profiles=af.CollectionPriorModel(gaussian=htf.profiles.Gaussian),
        settings=htf.SettingsPhase(),
        search=af.Emcee(),
    )

    """Note that we pass the info to the phase when we run it, so that the aggregator can make it accessible."""

    phase.run(dataset=dataset, mask=mask, info=info)

print("Emcee has finished run - you may now continue the notebook.")

# %%
"""
Checkout the output folder - you should see three new sets of results corresponding to our 3 Gaussian datasets.
Unlike previous tutorials, these folders in the output folder are named after the dataset and contain the folder
with the phase's name, as opposed to just the phase-name folder.

To load these results with the aggregator, we simply point it to the path of the results we want it to inspect.
"""

# %%
output_path = f"{workspace_path}/howtofit/output"

agg = af.Aggregator(directory=str(output_path))

# %%
"""
To begin, let me quickly explain what a generator is in Python, for those unaware. A generator is an object that 
iterates over a function when it is called. The aggregator creates all objects as generators, rather than lists, or 
dictionaries, or whatever.

Why? Because lists and dictionaries store every entry in memory simultaneously. If you fit many datasets, you'll have 
lots of results and therefore use a lot of memory. This will crash your laptop! On the other hand, a generator only 
stores the object in memory when it runs the function; it is free to overwrite it afterwards. Thus, your laptop won't 
crash!

There are two things to bare in mind with generators:

 1) A generator has no length, thus to determine how many entries of data it corresponds to you first must turn it to a 
 list.

 2) Once we use a generator, we cannot use it again - we'll need to remake it. For this reason, we typically avoid 
 storing the generator as a variable and instead use the aggregator to create them on use.

We can now create a 'samples' generator of every fit. An instance of the Samples class acts as an 
interface between the results of the non-linear fit on your hard-disk and Python.
"""

# %%
samples_gen = agg.values("samples")

# %%
"""
When we print this list of outputs you should see over 3 different MCMCSamples instances. There are more than 3
because the aggregator has loaded the results of previous tutorial as well as the 3 fits we performed above!
"""

# %%
print("Emcee Samples:\n")
print(samples_gen)
print("Total Samples Objects = ", len(list(samples_gen)), "\n")

# %%
"""
To remove the fits of previous tutorials and just keep the MCMCSamples of the 3 datasets fitted in this tutorial we can 
use the aggregator's filter tool. The phase name 'phase_t8' used in this tutorial is unique to all 3 fits, so we can 
use it to filter our results are desired.
"""

# %%
phase_name = "phase_t8"
agg_filter = agg.filter(agg.phase == phase_name)
samples_gen = agg_filter.values("samples")

# %%
"""
Alternatively, we can filter using strings, requiring that the string appears in the full path of the output
results. This is useful if you fit a large sample of data where:

 - Multiple results, corresponding to different phases and model-fits are stored in the same path.
 
 - Different runs using different _SettingsPhase_ are in the same path.
 
 - Fits using different non-linear searches, with different settings, are contained in the same path.

The example below shows us using the contains filter to get the results of the fit to only the first dataset. 
"""

# %%
agg_filter_contains = agg.filter(
    agg.directory.contains("phase_t8"), agg.directory.contains("gaussian_x1_0")
)
print("Directory Contains Filtered NestedSampler Samples: \n")
print(
    "Total Samples Objects = ", len(list(agg_filter_contains.values("samples"))), "\n\n"
)

# %%
"""
We've encountered the Samples class in previous tutorials, specifically the Result object returned from a phase. In 
this tutorial we'll inspect the Sampels class in more detail. The Samples class contains all the accepted parameter 
samples of the non-linear search, which is a list of lists where:

 - The outer list is the size of the total number of samples.
 
 - The inner list is the size of the number of free parameters in the fit.

"""

# %%
for samples in agg_filter.values("samples"):
    print("All parameters of the very first sample")
    print(samples.parameters[0])
    print("The tenth sample's third parameter")
    print(samples.parameters[9][2])

# %%
"""
The Samples class also contains the log likelihood, log prior, log posterior and weights of every accepted sample, 
where:

 - The log likelihood is the value evaluated from the likelihood function (e.g. -0.5 * chi_squared + the noise 
 normalized).

 - The log prior encodes information on how the priors on the parameters maps the log likelihood value to the log
 posterior value.

 - The log posterior is log_likelihood + log_prior.

 - The weight gives information on how samples should be combined to estimate the posterior. The weight values 
 depend on the sampler used, for MCMC samples they are all 1 (e.g. all weighted equally).

"""

# %%
for samples in agg_filter.values("samples"):
    print("log(likelihood), log(prior), log(posterior) and weight of the tenth sample.")
    print(samples.log_likelihoods[9])
    print(samples.log_priors[9])
    print(samples.log_posteriors[9])
    print(samples.weights[9])

# %%
"""
As expected, this list now has only 3 MCMCSamples, one for each dataset we fitted.
"""

# %%
print("Phase Name Filtered Emcee Samples:\n")
print(samples_gen)
print("Total Samples Objects = ", len(list(samples_gen)), "\n")

# %%
"""
We can use the samples to create a list of the maximum log likelihood model of each fit to our three images.
"""

# %%
vector = [samps.max_log_likelihood_vector for samps in agg_filter.values("samples")]
print("Maximum Log Likelihood Parameter Lists:\n")
print(vector, "\n")

# %%
"""
This provides us with lists of all model parameters. However, this isn't that much use - which values correspond
to which parameters?

Its more useful to create the maximum log likelihood model instance of every fit.
"""

# %%
instances = [
    samps.max_log_likelihood_instance for samps in agg_filter.values("samples")
]
print("Maximum Log Likelihood Model Instances:\n")
print(instances, "\n")

# %%
"""
A model instance contains all the model components of our fit - which for the fits above was a single gaussian
profile (the word 'gaussian' comes from what we called it in the CollectionPriorModel when making the phase above).
"""

# %%
print(instances[0].profiles.gaussian)
print(instances[1].profiles.gaussian)
print(instances[2].profiles.gaussian)

# %%
"""
This, of course, gives us access to any individual parameter of our maximum log likelihood model. Below, we see that t
he 3 Gaussians were simulated using sigma values of 1.0, 5.0 and 10.0.
"""

# %%
print(instances[0].profiles.gaussian.sigma)
print(instances[1].profiles.gaussian.sigma)
print(instances[2].profiles.gaussian.sigma)

# %%
"""
We can also access the 'most probable' model, which is the model computed by binning all of the accepted Emcee sample
parameters into a histogram, after removing the initial samples where the non-linear sampler is 'burning in' to the 
high likelihood regions of parameter space. 

The median of each 1D histogram (1 for each parameter) is then used to give the median PDF model. This process is 
called 'marginalization' and the hisograms which provide information on the probability estimates of each parameter 
are called the 'Probability Density Function' or PDF for short.
"""

# %%
mp_vectors = [samps.median_pdf_vector for samps in agg_filter.values("samples")]
mp_instances = [samps.median_pdf_instance for samps in agg_filter.values("samples")]

print("Most Probable Model Parameter Lists:\n")
print(mp_vectors, "\n")
print("Most probable Model Instances:\n")
print(mp_instances, "\n")

# %%
"""
We can compute the upper and lower errors on each parameter at a given sigma limit. These are computed via 
marginalization, whereby instead of using the median of the histogram (e.g. the parameter value at the 50% mark of the
histogram) the values at a specified sigma interval are used. For 3 sigma, these confidence intervals are at 0.3% and
99.7%.

Here, I use "ue3" to signify this is an upper error at 3 sigma confidence,, and "le3" for the lower error.
"""

# %%
ue3_vectors = [
    out.error_vector_at_upper_sigma(sigma=3.0) for out in agg_filter.values("samples")
]
ue3_instances = [
    out.error_instance_at_upper_sigma(sigma=3.0) for out in agg_filter.values("samples")
]
le3_vectors = [
    out.error_vector_at_lower_sigma(sigma=3.0) for out in agg_filter.values("samples")
]
le3_instances = [
    out.error_instance_at_lower_sigma(sigma=3.0) for out in agg_filter.values("samples")
]

print("Errors Lists:\n")
print(ue3_vectors, "\n")
print(le3_vectors, "\n")
print("Errors Instances:\n")
print(ue3_instances, "\n")
print(le3_instances, "\n")

# %%
"""
We can also print the "model_results" of all phases, which is string that summarizes every fit's model providing
quick inspection of all results.
"""

# %%
results = agg_filter.model_results
print("Model Results Summary:\n")
print(results, "\n")

# %%
"""
Lets end part 1 with something more ambitious. Lets create a plot of the inferred sigma values vs intensity of each
Gaussian profile, including error bars at 3 sigma confidence.
"""

# %%
import matplotlib.pyplot as plt

mp_instances = [samps.median_pdf_instance for samps in agg_filter.values("samples")]
ue3_instances = [
    out.error_instance_at_upper_sigma(sigma=3.0) for out in agg_filter.values("samples")
]
le3_instances = [
    out.error_instance_at_lower_sigma(sigma=3.0) for out in agg_filter.values("samples")
]

mp_sigmas = [instance.profiles.gaussian.sigma for instance in mp_instances]
ue3_sigmas = [instance.profiles.gaussian.sigma for instance in ue3_instances]
le3_sigmas = [instance.profiles.gaussian.sigma for instance in le3_instances]
mp_intensitys = [instance.profiles.gaussian.sigma for instance in mp_instances]
ue3_intensitys = [instance.profiles.gaussian.sigma for instance in ue3_instances]
le3_intensitys = [instance.profiles.gaussian.intensity for instance in le3_instances]

plt.errorbar(
    x=mp_sigmas,
    y=mp_intensitys,
    marker=".",
    linestyle="",
    xerr=[le3_sigmas, ue3_sigmas],
    yerr=[le3_intensitys, ue3_intensitys],
)
plt.show()



import numpy as np
import pytest

from autofit import message_passing as mp


@pytest.fixture(
    name="likelihood"
)
def make_likelihood(norm):
    def likelihood(z, y):
        return norm.logpdf(z - y)

    return likelihood


@pytest.fixture(
    name="model"
)
def make_model(
        likelihood_factor,
        linear_factor,
        prior_a,
        prior_b
):
    return likelihood_factor * linear_factor * prior_a * prior_b


@pytest.fixture(
    name="model_approx"
)
def make_model_approx(
        model
):
    a = np.array([[-1.3], [0.7]])
    b = np.array([-0.5])

    n_obs = 100
    n_features, n_dims = a.shape

    x = 5 * np.random.randn(n_obs, n_features)
    y = x.dot(a) + b + np.random.randn(n_obs, n_dims)

    message_a = mp.NormalMessage.from_mode(
        np.zeros((n_features, n_dims)),
        100
    )

    message_b = mp.NormalMessage.from_mode(
        np.zeros(n_dims),
        100
    )

    message_z = mp.NormalMessage.from_mode(
        np.zeros((n_obs, n_dims)),
        100
    )

    return mp.MeanFieldApproximation.from_kws(
        model,
        a=message_a,
        b=message_b,
        z=message_z,
        x=mp.FixedMessage(x),
        y=mp.FixedMessage(y))


def test_laplace(
        model_approx
):
    opt = mp.optimise.LaplaceOptimiser(
        model_approx,
        n_iter=3
    )
    opt.run()

    q_a = opt.model_approx['a']
    q_b = opt.model_approx['b']

    assert q_a.mu[0] == pytest.approx(-1.2, rel=1)
    assert q_a.sigma[0][0] == pytest.approx(0.04, rel=1)

    assert q_b.mu[0] == pytest.approx(-0.5, rel=1)
    assert q_b.sigma[0] == pytest.approx(0.2, rel=1)


def test_importance_sampling(
        model,
        model_approx
):
    sampler = mp.ImportanceSampler(n_samples=500)

    history = {}
    n_iter = 3

    for i in range(n_iter):
        for factor in model.factors:
            # We have reduced the entire EP step into a single function
            model_approx, status = mp.sampling.project_model(
                model_approx,
                factor,
                sampler,
                force_sample=False,
                delta=1.
            )

            # save and print current approximation
            history[i, factor] = model_approx

    q_a = model_approx['a']
    q_b = model_approx['b']

    assert q_a.mu[0] == pytest.approx(-1.2, rel=1)
    assert q_a.sigma[0][0] == pytest.approx(7.13, rel=1)

    assert q_b.mu[0] == pytest.approx(-0.5, rel=1)
    assert q_b.sigma[0] == pytest.approx(6.8, rel=1)



import autofit as af

from howtofit.chapter_1_introduction.tutorial_6_complex_models.src.fit import fit as f
from howtofit.chapter_1_introduction.tutorial_6_complex_models.src.phase import (
    visualizer,
)

"""
The 'analysis.py' module in this tutorial is changed from tutorial 5, such that the fit function now assumes multiple
profiles are passed to it rather than a single Gaussian, see the comments below!
"""


class Analysis(af.Analysis):
    def __init__(self, masked_dataset, image_path=None):

        super().__init__()

        """The masked dataset and visualizer are created in the same way as tutorial 5."""

        self.masked_dataset = masked_dataset

        self.visualizer = visualizer.Visualizer(
            masked_dataset=self.masked_dataset, image_path=image_path
        )

    def log_likelihood_function(self, instance):
        """Determine the fit of a list of Profiles (Gaussians, Exponentials, etc.) to the dataset, using a
        model instance.

        Parameters
        ----------
        instance
            The list of Profile model instance (e.g. the Gaussians, Exponentials, etc.).

        Returns
        -------
        fit : Fit.log_likelihood
            The log likelihood value indicating how well this model fit the masked dataset.
        """

        """
        In tutorials 3 & 4, the instance was an instance of a single Gaussian profile. PyAutoFit knew this instance
        would contain just one Gaussian, because when the phase was created we used a _PriorModel_ object in PyAutoFit
        to make the Gaussian. This meant we could create the model data using the line:

            model_data = instance.gaussian.profile_from_xvalues(xvalues=self.masked_dataset.xvalues)

        In this tutorial our instance is comprised of multiple Profile objects. This is reflected the main tutorial
        script where we create a model using a CollectionPriorModel:

            model = CollectionPriorModel(gaussian=profiles.Gaussian, exponential=profiles.Exponential).

        By using a CollectionPriorModel, this means the instance parameter input into the fit function is a
        dictionary where individual profiles (and their parameters) can be accessed as followed:
        
            print(instance.profiles.gaussian)
            print(instance.profiles.exponential)
            print(instance.profiles.exponential.centre)

        The names of the attributes of the instance correspond to what we input into the CollectionPriorModel. Lets
        look at a second example:

            model = CollectionPriorModel(
                          gaussian_0=profiles.Gaussian,
                          gaussian_1=profiles.Gaussian,
                          whatever_i_want=profiles.Exponential
                     ).

            print(instance.profiles.gaussian_0)
            print(instance.profiles.gaussian_1)
            print(instance.profiles.whatever_i_want.centre)

        A CollectionPriorModel allows us to name our model components whatever we want!

        In this tutorial, we want our 'fit' function to fit the data with a profile which is the summed profile
        of all individual profiles in the model. Look at 'model_data_from_instance' to see how we do this.
        """

        model_data = self.model_data_from_instance(instance=instance)
        fit = self.fit_from_model_data(model_data=model_data)
        return fit.log_likelihood

    def model_data_from_instance(self, instance):

        """
        To create the summed profile of all individual profiles in an instance, we can use a list comprehension
        to iterate over all profiles in the instance.
        """

        return sum(
            [
                profile.profile_from_xvalues(xvalues=self.masked_dataset.xvalues)
                for profile in instance.profiles
            ]
        )

        """
        For those not familiar with list comprehensions, below I've included how one would use the instance to create
        the summed profile using a more simple for loop.
        
            model_data = np.zeros(shape=self.masked_dataset.xvalues.shape[0])

            for profile in instance:
                model_data += profile.profile_from_xvalues(xvalues=self.masked_dataset.xvalues)
    
            return model_data
        """

    def fit_from_model_data(self, model_data):
        return f.FitDataset(masked_dataset=self.masked_dataset, model_data=model_data)

    def visualize(self, instance, during_analysis):

        """
        It is worth noting here why we create specific methods for creating the model_data and fit in an analysis.
        By doing so, the code in our visualize function (and also in the 'result.py' module) do not need changing
        even though we are now using a model with more components, requiring to sum their individual profiles.
        """

        model_data = self.model_data_from_instance(instance=instance)
        fit = self.fit_from_model_data(model_data=model_data)

        self.visualizer.visualize_fit(fit=fit, during_analysis=during_analysis)



# %%
"""
Tutorial 6: Complex Models
==========================

Up to now, we've fitted a very simple model - a single 1D Gaussian with just 3 free parameters. In this tutorial,
we'll look at how PyAutoFit allows us to compose and fit models of arbitrary complexity.

To begin, you should check out the module 'tutorial_6_complex_models/model/profiles.py'. In previous tutorials this
module was called 'gaussian.py' and it contained only the Gaussian class we fitted to data. The module now includes
a second profile, 'Exponential', which like the Gaussian class is a model-component that can be fitted to data.

Up to now, our data has always been generated using a single Gaussian profile. Thus, we have only needed to fit
it with a single Gaussian. In this tutorial, our dataset is now a superpositions of multiple profiles (e.g a
Gaussians + Exponential). The models we compose and fit are therefore composed of multiple profiles, such that when we
generate the model-data we generate it as the sum of all individual profiles in our model.
"""

# %%
#%matplotlib inline

from autoconf import conf
import autofit as af
from howtofit.chapter_1_introduction.tutorial_6_complex_models import src as htf

import numpy as np

from pyprojroot import here

workspace_path = str(here())
print("Workspace Path: ", workspace_path)

# %%
"""
Setup the configs as we did in the previous tutorial, as well as the output folder for our non-linear search.
"""

# %%
conf.instance = conf.Config(
    config_path=f"{workspace_path}/howtofit/config",
    output_path=f"{workspace_path}/howtofit/output",
)

# %%
"""
Lets quickly recap tutorial 1, where using PriorModels we created a Gaussian as a model component and used it to map a
list of parameters to a model instance.
"""

# %%
model = af.PriorModel(htf.profiles.Gaussian)

print("PriorModel Gaussian object: \n")
print(model)

instance = model.instance_from_vector(vector=[0.1, 0.2, 0.3])

print("Model Instance: \n")
print(instance)

print("Instance Parameters \n")
print("x = ", instance.centre)
print("intensity = ", instance.intensity)
print("sigma = ", instance.sigma)

# %%
"""
Defining a model using multiple model components is straight forward in PyAutoFit, using a CollectionPriorModel
object.
"""

# %%
model = af.CollectionPriorModel(
    gaussian=htf.profiles.Gaussian, exponential=htf.profiles.Exponential
)

# %%
"""
A CollectionPriorModel behaves like a _PriorModel_ but contains a collection of model components. For example, it can
create a model instance by mapping a list of parameters, which in this case is 6 (3 for the Gaussian [centre,
intensity, sigma] and 3 for the Exponential [centre, intensity, rate]).
"""

# %%
instance = model.instance_from_vector(vector=[0.1, 0.2, 0.3, 0.4, 0.5, 0.01])

# %%
"""
This instance contains each of the model components we defined above, using the input argument name of the
CollectionoPriorModel to define the attributes in the instance:
"""

# %%
print("Instance Parameters \n")
print("x (Gaussian) = ", instance.gaussian.centre)
print("intensity (Gaussian) = ", instance.gaussian.intensity)
print("sigma (Gaussian) = ", instance.gaussian.sigma)
print("x (Exponential) = ", instance.exponential.centre)
print("intensity (Exponential) = ", instance.exponential.intensity)
print("sigma (Exponential) = ", instance.exponential.rate)

# %%
"""
We can call the components of a CollectionPriorModel whatever we want, and the mapped instance will use those names.
"""

# %%
model_custom_names = af.CollectionPriorModel(
    james=htf.profiles.Gaussian, rich=htf.profiles.Exponential
)

instance = model_custom_names.instance_from_vector(
    vector=[0.1, 0.2, 0.3, 0.4, 0.5, 0.01]
)

print("Instance Parameters \n")
print("x (Gaussian) = ", instance.james.centre)
print("intensity (Gaussian) = ", instance.james.intensity)
print("sigma (Gaussian) = ", instance.james.sigma)
print("x (Exponential) = ", instance.rich.centre)
print("intensity (Exponential) = ", instance.rich.intensity)
print("sigma (Exponential) = ", instance.rich.rate)

# %%
"""
Now we can create a model composed of multiple components, lets fit it to a dataset. To do this, we updated this 
tutorial's phase package, spefically its 'Analysis' class such that it creates model data as a super position of all of 
the model's individual profiles. For example, in the model above, the model data is the sum of the Gaussian's 
individual profile and Exponential's individual profile.

Checkout 'phase.py' and 'analysis.py' now, for a description of how this has been implemented.
"""

# %%
"""
Import the simulator module and set up the dataset. This uses a new dataset that is generated as a sum of a Gaussian 
and Exponential profile.
"""

# %%
from howtofit.simulators.chapter_1 import gaussian_x1_exponential_x1

dataset = htf.Dataset(
    data=gaussian_x1_exponential_x1.data, noise_map=gaussian_x1_exponential_x1.noise_map
)

# %%
"""
We again need to create a mask for our data. In this exmample, we'll omit actual masking of the dataset, but still
need to define a mask to pass the 'phase.run' method.
"""

# %%
mask = np.full(fill_value=False, shape=dataset.data.shape)

# %%
"""
Lets now perform the fit using our model which is composed of two profiles. You'll note that the Emcee
dimensionality has increased from N=3 to N=6, given that we are now fitting two profiles each with 3 free parameters.
"""

# %%
phase = htf.Phase(
    phase_name="phase_t6_gaussian_x1_exponential_x1",
    profiles=af.CollectionPriorModel(
        gaussian=htf.profiles.Gaussian, exponential=htf.profiles.Exponential
    ),
    search=af.Emcee(),
)

print(
    "Emcee has begun running - checkout the autofit_workspace/howtofit/chapter_1_introduction/output/phase_t5_gaussian_x1_exponential_x1"
    "folder for live output of the results."
    "This Jupyter notebook cell with progress once Emcee has completed - this could take a few minutes!"
)

phase.run(dataset=dataset, mask=mask)

print("Emcee has finished run - you may now continue the notebook.")

# %%
"""
Quickly inspect the results of this fit, which you may have noticed takes a bit longer to run than the fits performed
in previous tutorials. This is because the dimensionality of the model we are fitted increased from 3 to 6.

With the CollectionPriorModel, PyAutoFit gives us all the tools we need to compose and fit any model imaginable!
Lets fit a model composed of two Gaussians and and an Exponential, which will have a dimensionality of N=9.
"""

# %%
from howtofit.simulators.chapter_1 import gaussian_x2_exponential_x1

dataset = htf.Dataset(
    data=gaussian_x2_exponential_x1.data, noise_map=gaussian_x2_exponential_x1.noise_map
)

phase = htf.Phase(
    phase_name="phase_t6_gaussian_x2_exponential_x1",
    profiles=af.CollectionPriorModel(
        gaussian_0=htf.profiles.Gaussian,
        gaussian_1=htf.profiles.Gaussian,
        exponential=htf.profiles.Exponential,
    ),
    search=af.Emcee(),
)

print(
    "Emcee has begun running - checkout the autofit_workspace/howtofit/chapter_1_introduction/output/phase_t5_gaussian_x2_exponential_x1"
    "folder for live output of the results."
    "This Jupyter notebook cell with progress once Emcee has completed - this could take a few minutes!"
)

phase.run(dataset=dataset, mask=mask)

print("Emcee has finished run - you may now continue the notebook.")

# %%
"""
We can fully custommize the model that we fit. Lets suppose we have a dataset that consists of three Gaussian line
profiles, but we also know the following information about the dataset:

- All 3 Gaussians are centrally aligned.
- The sigma of one Gaussian is equal to 1.0.

We can edit our CollectionPriorModel to meet these constraints accordingly:
"""

# %%
model = af.CollectionPriorModel(
    gaussian_0=htf.profiles.Gaussian,
    gaussian_1=htf.profiles.Gaussian,
    gaussian_2=htf.profiles.Gaussian,
)

# %%
"""
This aligns the centres of the 3 Gaussians, reducing the dimensionality of the model from N=9 to N=7
"""

# %%
model.gaussian_0.centre = model.gaussian_1.centre
model.gaussian_1.centre = model.gaussian_2.centre

# %%
"""
This fixes the sigma value of one Gaussian to 1.0, further reducing the dimensionality from N=7 to N=6.
"""

# %%
model.gaussian_0.sigma = 1.0

# %%
"""
We can now fit this model using a phase as per usual.
"""

# %%
from howtofit.simulators.chapter_1 import gaussian_x3

dataset = htf.Dataset(data=gaussian_x3.data, noise_map=gaussian_x3.noise_map)

phase = htf.Phase(phase_name="phase_t6_gaussian_x3", profiles=model, search=af.Emcee())

print(
    "Emcee has begun running - checkout the autofit_workspace/howtofit/chapter_1_introduction/output/phase_t5_gaussian_x3"
    "folder for live output of the results."
    "This Jupyter notebook cell with progress once Emcee has completed - this could take a few minutes!"
)

phase.run(dataset=dataset, mask=mask)

print("Emcee has finished run - you may now continue the notebook.")

# %%
"""
And with that, we are complete. In this tutorial, we learned how to compose complex models in PyAutoFit and adjust our
'phase.py' and 'analyis.py' modules to fit them. To end, you should think again in more detail about your model
fitting problem:

Are there many different model components you may wish to define and fit?
Is your model-data the super position of many different model components, like the profiles in this tutorial?

In this tutorial, all components of our model did the same thing - represent a 'line' of data. In your model, you may
have model components that represent different parts of your model, which need to be combined in more complicated ways
in order to create your model-fit. In such circumstances, the 'fit' method in your 'Analysis' class may be significantly
more complex than the example shown in this tutorial. Nevertheless, you now have all the tools you need to define,
compose and fit very complex models - there isn't much left for you to learn on your journey through PyAutoFit!
"""



import autofit as af

# The 'settings.py' module is identical to the previous tutorial.


class SettingsPhase(af.AbstractSettingsPhase):
    def __init__(self, data_trim_left=None, data_trim_right=None):

        super().__init__()

        self.data_trim_left = data_trim_left
        self.data_trim_right = data_trim_right

    @property
    def tag(self):

        # You may well have many more tags which appear here.

        return (
            "settings"  # For every tag you add, you'll add it to this return statement
            + self.data_trim_left_tag
            + self.data_trim_right_tag
            # e.g. + your_own_tag
        )

    # This function generates a string we'll use to 'tag' a phase which uses this setting, thus ensuring results are
    # output to a unique path.

    @property
    def data_trim_left_tag(self):
        """Generate a data trim left tag, to customize phase names based on how much of the dataset is trimmed to 
        its left.
    
        This changes the phase name 'settings' as follows:
    
        data_trim_left = None -> settings
        data_trim_left = 2 -> settings__trim_left_2
        data_trim_left = 10 -> settings__trim_left_10
        """
        if self.data_trim_left is None:
            return ""
        return "__trim_left_" + str(self.data_trim_left)

    @property
    def data_trim_right_tag(self):
        """Generate a data trim right tag, to customize phase names based on how much of the dataset is trimmed to its right.
    
        This changes the phase name 'settings' as follows:
    
        data_trim_right = None -> settings
        data_trim_right = 2 -> settings__trim_right_2
        data_trim_right = 10 -> settings__trim_right_10
        """
        if self.data_trim_right is None:
            return ""
        return "__trim_right_" + str(self.data_trim_right)



from howtofit.chapter_1_introduction.tutorial_7_phase_customization.src.dataset import (
    dataset as ds,
)

"""
In this tutorial, we run phases where the dataset we input into the phase is altered before the model-fitting
procedure is run. The dataset is trimmed by an input number of pixels to the left and / or right.

The 'meta_dataset.py' module is the module in PyAutoFit which handles the creation of these new datasets. If we want
to have the option of a phase editing the data-set, the parameters which control this (e.g. 'data_trim_left')
are stored here and then used when the 'phase.run' method is called.

Your model-fitting problem may not require the meta_dataset.py module. If so that is fine, and you can revert to the
templates of the previous tutorials which do not use one. It really depends on the nature of your problem.
"""


class MetaDataset:

    """
    The data_trim_left and data_trim_right are passed to the phase when it is set up and stored in an
    instance of the 'MetaDataset' class.
    """

    def __init__(self, settings):
        """An intermediate meta-dataset class which is used to create the masked dataset, but with phase-dependent
        settings applied that augment the data. In this example, the left and right edges of the data and mask can be
        trimmed.

        Parameters
        ----------
        settings : SettingsPhase
            The collection of settings of the phase used to augment the data that is fitted and tag the output path.
        """

        self.settings = settings

    """
    The masked dataset that is fitted by an analysis is created by the MetaDataset class using the method below.

    If the MetaDataset's data trim attributes are not None, they are used to trim the masked-dataset before it is
    fitted.
    """

    def masked_dataset_from_dataset_and_mask(self, dataset, mask):
        """Create a masked dataset from the input dataset and a mask.

        The dataset will have its left and / or right edge trimmed and removed from the model-fit, if the phase
        settings are input to do this.

        Parameters
        ----------
        dataset : dataset.Dataset
            The dataset that is masked, trimmed and fitted.
        mask : ndarray
            The mask applied to the dataset, which is also trimmed.
        """

        masked_dataset = ds.MaskedDataset(dataset=dataset, mask=mask)

        if self.settings.data_trim_left is not None:
            masked_dataset = masked_dataset.with_left_trimmed(
                data_trim_left=self.settings.data_trim_left
            )

        if self.settings.data_trim_right is not None:
            masked_dataset = masked_dataset.with_right_trimmed(
                data_trim_right=self.settings.data_trim_right
            )

        return masked_dataset



# %%
"""
Tutorial 8: Aggregator Part 2
=============================

In part 1 of tutorial 8, we fitted 3 datasets and used the aggregator to load their results. We focused on the
results of the non-linear search, Emcee. In part 2, we'll look at how the way we designed our source code
makes it easy to use these results to plot results and data.
"""

# %%
#%matplotlib inline

from autoconf import conf
import autofit as af
from howtofit.chapter_1_introduction.tutorial_8_aggregator import src as htf

from pyprojroot import here

workspace_path = str(here())
print("Workspace Path: ", workspace_path)

# %%
"""
Setup the configs as we did in the previous tutorial, as well as the output folder for our non-linear search.
"""

# %%
conf.instance = conf.Config(
    config_path=f"{workspace_path}/howtofit/config",
    output_path=f"{workspace_path}/howtofit/output",
)

# %%
"""
To load these results with the aggregator, we again point it to the path of the results we want it to inspect.
"""

# %%
output_path = f"{workspace_path}/howtofit/output"

agg = af.Aggregator(directory=str(output_path))
phase_name = "phase_t8"
agg_filter = agg.filter(agg.phase == phase_name)

# %%
"""
We can use the aggregator to load a generator of every fit's dataset, by changing the 'output' attribute to the 
'dataset' attribute at the end of the aggregator. We'll filter by phase name again to get datasets of only the fits 
performed for this tutorial.

Note that we had to manually specify in the 'phase.py' for the dataset to be saved too hard-disk such that the 
aggregator can load it.
"""

# %%
dataset_gen = agg_filter.values("dataset")
print("Datasets:")
print(list(dataset_gen), "\n")

# %%
"""
It is here the object-based design of our plot module comes into its own. We have the dataset objects loaded, meaning
we can easily plot each dataset using the 'dataset_plot.py' module.
"""

# %%
for dataset in agg_filter.values("dataset"):
    htf.plot.Dataset.data(dataset=dataset)

# %%
"""
The dataset names are available, either as part of the dataset or via the aggregator's dataset_names method.
"""

# %%
for dataset in agg_filter.values("dataset"):
    print(dataset.name)

# %%
"""
The info dictionary we input into the pipeline is also available.
"""

# %%
for info in agg_filter.values("info"):
    print(info)

# %%
"""
We can repeat the same trick to get the mask of every fit.
"""

# %%
mask_gen = agg_filter.values("mask")
print("Masks:")
print(list(mask_gen), "\n")


# %%
"""
We're going to refer to our datasets using the best-fit model of each phase. To do this, we'll need each phase's masked
dataset.

(If you are unsure what the 'zip' is doing below, it essentially combines the'datasets' and 'masks' lists in such
a way that we can iterate over the two simultaneously to create each MaskedDataset).

The masked dataset may have been altered by the data_trim_ custom phase settings. We can load the meta_dataset via the 
aggregator to use these settings when we create the masked dataset.
"""

# %%
dataset_gen = agg_filter.values("dataset")
mask_gen = agg_filter.values("mask")

masked_datasets = [
    htf.MaskedDataset(dataset=dataset, mask=mask)
    for dataset, mask in zip(dataset_gen, mask_gen)
]

masked_datasets = [
    masked_dataset.with_left_trimmed(data_trim_left=meta_dataset.data_trim_left)
    for masked_dataset, meta_dataset in zip(
        masked_datasets, agg_filter.values("meta_dataset")
    )
]
masked_datasets = [
    masked_dataset.with_right_trimmed(data_trim_right=meta_dataset.data_trim_right)
    for masked_dataset, meta_dataset in zip(
        masked_datasets, agg_filter.values("meta_dataset")
    )
]

# %%
"""
There is a problem with how we set up the masked datasets above, can you guess what it is?

We used lists! If we had fit a large sample of data, the above object would store the masked dataset of all objects
simultaneously in memory on our hard-disk, likely crashing our laptop! To avoid this, we must write functions that
manipulate the aggregator generators as generator themselves. Below is an example function that performs the same
task as above.
"""

# %%
def masked_dataset_from_agg_obj(agg_obj):

    dataset = agg_obj.dataset
    mask = agg_obj.mask

    masked_dataset = htf.MaskedDataset(dataset=dataset, mask=mask)

    meta_dataset = agg_obj.meta_dataset

    masked_dataset = masked_dataset.with_left_trimmed(
        data_trim_left=meta_dataset.data_trim_left
    )
    masked_dataset = masked_dataset.with_right_trimmed(
        data_trim_right=meta_dataset.data_trim_right
    )

    return masked_dataset


# %%
"""
To manipulate this function as a generator using the aggregator, we must apply it to the aggregator's map function.

The masked_dataset_generator below ensures that we avoid representing all masked datasets simultaneously in memory.
"""

# %%
masked_dataset_gen = agg_filter.map(func=masked_dataset_from_agg_obj)
print(list(masked_dataset_gen))

# %%
"""
Lets get the the maximum likelihood model instances, as we did in part 1.
"""

# %%
instances = [
    samps.max_log_likelihood_instance for samps in agg_filter.values("samples")
]

# %%
"""
Okay, we want to inspect the fit of each best-fit model. To do this, we reperform each fit.

First, we need to create the model-data of every best-fit model instance. Lets begin by creating a list of profiles of
every phase.
"""

# %%
profiles = [instance.profiles for instance in instances]

# %%
"""
We can use these to create the model data of each set of profiles (Which in this case is just 1 Gaussian, but had
we included more profiles in the model would consist of multiple Gaussians / Exponentials).
"""

# %%
model_datas = [
    profile.gaussian.profile_from_xvalues(xvalues=dataset.xvalues)
    for profile, dataset in zip(profiles, agg_filter.values("dataset"))
]

# %%
"""
And, as we did in tutorial 2, we can combine the masked_datasets and model_datas in a Fit object to create the
maximum likelihood fit of each phase!
"""

# %%
fits = [
    htf.FitDataset(masked_dataset=masked_dataset, model_data=model_data)
    for masked_dataset, model_data in zip(masked_datasets, model_datas)
]

# %%
"""
We can now plot different components of the fit (again benefiting from how we set up the 'fit_plots.py' module)!
"""

# %%
for fit in fits:
    htf.plot.FitDataset.residual_map(fit=fit)
    htf.plot.FitDataset.normalized_residual_map(fit=fit)
    htf.plot.FitDataset.chi_squared_map(fit=fit)

# %%
"""
Again, the code above does not use generators and could prove memory intensive for large datasets. Below is how we 
would perform the above task with generator functions, using the masked_dataset_generator above for the masked 
dataset.
"""

# %%
def model_data_from_agg_obj(agg_obj):
    xvalues = agg_obj.dataset.xvalues
    instance = agg_obj.samples.max_log_likelihood_instance
    profiles = instance.profiles
    model_data = sum(
        [profile.profile_from_xvalues(xvalues=xvalues) for profile in profiles]
    )

    return model_data


def fit_from_agg_obj(agg_obj):
    masked_dataset = masked_dataset_from_agg_obj(agg_obj=agg_obj)
    model_data = model_data_from_agg_obj(agg_obj=agg_obj)

    return htf.FitDataset(masked_dataset=masked_dataset, model_data=model_data)


fit_gen = agg_filter.map(func=fit_from_agg_obj)

for fit in fit_gen:
    htf.plot.FitDataset.residual_map(fit=fit)
    htf.plot.FitDataset.normalized_residual_map(fit=fit)
    htf.plot.FitDataset.chi_squared_map(fit=fit)

# %%
"""
Setting up the above objects (the masked_datasets, model datas, fits) was a bit of work. It wasn't too many
lines of code, but for something we'll likely want to do many times it'd be nice to have a short cut to setting
them up, right?

In 'aggregator.py' we've set up exactly such a short-cut. This module simply contains the generator functions above 
such that the generator can be created by passing the aggregator. This provides us with convenience methods for quickly 
creating the masked dataset, model data and fits using single lines of code:
"""

# %%

from howtofit.chapter_1_introduction.tutorial_8_aggregator.src.phase import aggregator

masked_dataset_gen = aggregator.masked_dataset_generator_from_aggregator(
    aggregator=agg_filter
)
model_data_gen = aggregator.model_data_generator_from_aggregator(aggregator=agg_filter)
fit_gen = aggregator.fit_generator_from_aggregator(aggregator=agg_filter)

# %%
"""
For your model-fitting project, you'll need to update the 'aggregator.py' module in the same way. This is why we have 
emphasised the object-oriented design of our model-fitting project through. This design makes it very easy to inspect 
results via the aggregator later on!
"""



from howtofit.chapter_1_introduction.tutorial_8_aggregator.src.dataset import (
    dataset as ds,
)

# The 'meta_dataset.py' module is unchanged from the previous tutorial.


class MetaDataset:
    def __init__(self, settings):
        """An intermediate meta-dataset class which is used to create the masked dataset, but with phase-dependent
        settings applied that augment the data. In this example, the left and right edges of the data and mask can be
        trimmed.

        Parameters
        ----------
        settings : SettingsPhase
            The collection of settings of the phase used to augment the data that is fitted and tag the output path.
        """

        self.settings = settings

    def masked_dataset_from_dataset_and_mask(self, dataset, mask):
        """Create a masked dataset from the input dataset and a mask.

        The dataset will have its left and / or right edge trimmed and removed from the model-fit, if the phase
        settings are input to do this.

        Parameters
        ----------
        dataset : dataset.Dataset
            The dataset that is masked, trimmed and fitted.
        mask : ndarray
            The mask applied to the dataset, which is also trimmed.
        """

        masked_dataset = ds.MaskedDataset(dataset=dataset, mask=mask)

        if self.settings.data_trim_left is not None:
            masked_dataset = masked_dataset.with_left_trimmed(
                data_trim_left=self.settings.data_trim_left
            )

        if self.settings.data_trim_right is not None:
            masked_dataset = masked_dataset.with_right_trimmed(
                data_trim_right=self.settings.data_trim_right
            )

        return masked_dataset



import numpy as np

from autofit import message_passing as mp


def func(x):
    return 2 * x


def test_():
    x = mp.Variable("x")
    factor = mp.Factor(
        func
    )(x)

    model = factor * factor

    value = model(np.array([1.0]))
    assert value.log_value == np.array([4.0])



import numpy as np

import autofit as af
from autofit import message_passing as mp
from .model import Gaussian, make_data, _gaussian, _likelihood

n_observations = 100

prior = af.GaussianPrior(
    mean=0,
    sigma=40
)


def test_gaussian():
    x, y = make_data(
        Gaussian(
            centre=50.0,
            intensity=25.0,
            sigma=10.0
        ),
        n_observations
    )

    observations = mp.Plate(
        name="observations"
    )

    # TODO: Can we derive variables by looking at function argument names?
    x_ = mp.Variable(
        "x", observations
    )
    y_ = mp.Variable(
        "y", observations
    )
    z = mp.Variable(
        "z", observations
    )
    centre = mp.Variable(
        "centre"
    )
    intensity = mp.Variable(
        "intensity"
    )
    sigma = mp.Variable(
        "sigma"
    )

    gaussian = mp.Factor(
        _gaussian
    )(
        x_,
        centre,
        intensity,
        sigma
    ) == z
    likelihood = mp.Factor(
        _likelihood
    )(z, y_)

    # TODO: Can priors look like autofit priors? Could mp objects derive promise functionality from autofit?
    prior_centre = mp.Factor(
        prior
    )(centre)
    prior_intensity = mp.Factor(
        prior
    )(intensity)
    prior_sigma = mp.Factor(
        prior
    )(sigma)

    model = likelihood * gaussian * prior_centre * prior_sigma * prior_intensity

    model_approx = mp.MeanFieldApproximation.from_kws(
        model,
        centre=mp.NormalMessage.from_prior(
            prior
        ),
        intensity=mp.NormalMessage.from_prior(
            prior
        ),
        sigma=mp.NormalMessage.from_prior(
            prior
        ),
        x=mp.FixedMessage(x),
        y=mp.FixedMessage(y),
        z=mp.NormalMessage.from_mode(
            np.zeros(n_observations), 100
        ),
    )

    opt = mp.optimise.LaplaceOptimiser(
        model_approx,
        n_iter=3
    )
    opt.run()

    for string in ("centre", "intensity", "sigma"):
        print(f"{string} = {opt.model_approx[string].mu}")



import pytest
from scipy import stats


from autofit import message_passing as mp


@pytest.fixture(
    name="x"
)
def make_x():
    return mp.Variable("x")


@pytest.fixture(
    name="probit_factor"
)
def make_probit_factor(x):
    return mp.Factor(
        stats.norm(
            loc=0.,
            scale=1.
        ).logcdf
    )(x)



import numpy as np

"""
This is our dataset, which in tutorial 1 stores a 1D Gaussian with a noise-map.

Our data is a 1D numpy array of values corresponding to the observed counts of the Gaussian.
The noise-map corresponds to the expected noise in every data point.

In PyAutoFit we use the term 'dataset' to represent the collection of data, noise-map, etc. that are fitted. In this
chapter our dataset is a 1D line, for your model-fitting problem your dataset is probably more complex than this.

In these tutorials we are calling our dataset class 'Dataset', as opposed to something more specific to the data
(e.g. 'Line'). We advise that the dataset class is give a more specific name (e.g., if your dataset is imaging data,
the class may be called 'Imaging'. If your dataset is a spectrum, you may use 'Spectrum').

However, to make the tutorials clear, we've stuck with the name 'Dataset' for clarity. The example project at the
end of the chapter will adopt a more specific dataset naming convention.
"""


class Dataset:
    def __init__(self, data, noise_map):
        """A class containing the data and noise-map of a 1D line dataset.

        Parameters
        ----------
        data : np.ndarray
            The array of the data, in arbitrary units.
        noise_map : np.ndarray
            An array describing the RMS standard deviation error in each data pixel, in arbitrary units.
        """
        self.data = data
        self.noise_map = noise_map

    """The 'xvalues' defines the x coordinates of the 1D Gaussian, which we assume run from 0 the length of the data."""

    @property
    def xvalues(self):
        return np.arange(self.data.shape[0])



from abc import ABC, abstractmethod
from functools import reduce
from itertools import chain
from operator import and_
from typing import Optional, Tuple

import numpy as np


class AbstractMessage(ABC):
    log_base_measure: float

    _parameter_support: Optional[Tuple[Tuple[float, float], ...]] = None
    _support: Optional[Tuple[Tuple[float, float], ...]] = None

    @property
    @abstractmethod
    def natural_parameters(self):
        pass

    @abstractmethod
    def sample(self, n_samples):
        pass

    @staticmethod
    @abstractmethod
    def invert_natural_parameters(natural_parameters):
        pass

    @staticmethod
    @abstractmethod
    def to_canonical_form(x):
        pass

    @property
    @abstractmethod
    def log_partition(self):
        pass

    @property
    @abstractmethod
    def mean(self):
        pass

    @property
    @abstractmethod
    def variance(self):
        pass

    @property
    def scale(self):
        return self.variance ** 0.5

    def __init__(self, parameters, log_norm=0.):
        self.log_norm = log_norm
        self._broadcast = np.broadcast(*parameters)
        self.parameters = parameters

    def __iter__(self):
        return iter(self.parameters)

    @property
    def shape(self):
        return self._broadcast.shape

    @property
    def ndim(self):
        return self._broadcast.ndim

    @classmethod
    def from_natural_parameters(cls, parameters, **kwargs):
        args = cls.invert_natural_parameters(parameters)
        return cls(*args, **kwargs)

    @classmethod
    @abstractmethod
    def invert_sufficient_statistics(cls, sufficient_statistics):
        pass

    @classmethod
    def from_sufficient_statistics(cls, suff_stats, **kwargs):
        natural_params = cls.invert_sufficient_statistics(suff_stats)
        return cls.from_natural_parameters(natural_params, **kwargs)

    def sum_natural_parameters(self, *dists):
        """return the unnormalised result of multiplying the pdf
        of this distribution with another distribution of the same
        type
        """
        log_norm = self.log_norm + sum(
            dist.log_norm for dist in self._iter_dists(dists))
        new_params = sum(
            (dist.natural_parameters for dist in self._iter_dists(dists)),
            self.natural_parameters)
        mul_dist = self.from_natural_parameters(new_params, log_norm=log_norm)
        return mul_dist

    def sub_natural_parameters(self, other):
        """return the unnormalised result of dividing the pdf
        of this distribution with another distribution of the same
        type"""
        log_norm = self.log_norm - other.log_norm
        new_params = (self.natural_parameters - other.natural_parameters)
        div_dist = self.from_natural_parameters(new_params, log_norm=log_norm)
        return div_dist

    _multiply = sum_natural_parameters
    _divide = sub_natural_parameters

    def __mul__(self, other):
        if np.isscalar(other):
            log_norm = self.log_norm + np.log(other)
            return type(self)(*self.parameters, log_norm=log_norm)
        else:
            return self._multiply(other)

    def __rmul__(self, other):
        return self * other

    def __truediv__(self, other):
        if np.isscalar(other):
            log_norm = self.log_norm - np.log(other)
            return type(self)(*self.parameters, log_norm=log_norm)
        else:
            return self._divide(other)

    def __pow__(self, other):
        natural = self.natural_parameters
        new_params = other * natural
        log_norm = other * self.log_norm
        return self.from_natural_parameters(new_params, log_norm=log_norm)

    def __repr__(self):
        return f"{type(self).__name__}({self}"

    def __str__(self):
        return ""

    def logpdf(self, x):
        if np.shape(x) == self.shape:
            eta = self.natural_parameters
            t = self.to_canonical_form(x)
            log_base = self.log_base_measure
            eta_t = (eta * t).sum(0)  # TODO this can be made more efficient using tensordot
            return log_base + eta_t - self.log_partition
        elif np.shape(x)[1:] == self.shape:
            return np.array([self.logpdf(x_) for x_ in x])

        raise ValueError(
            f"shape of passed value {x.shape} does not match message shape {self.shape}")

    def pdf(self, x):
        return np.exp(self.logpdf(x))

    @classmethod
    def project(cls, samples, log_weights):
        """Calculates the sufficient statistics of a set of samples
        and returns the distribution with the appropriate parameters
        that match the sufficient statistics
        """
        # if weights aren't passed then equally weight all samples

        # Numerically stable weighting for very small/large weights
        log_w_max = np.max(log_weights, axis=0, keepdims=True)
        w = np.exp(log_weights - log_w_max)
        norm = w.mean(0)
        log_norm = np.log(norm) + log_w_max[0]

        tx = cls.to_canonical_form(samples)
        w /= norm
        suff_stats = (tx * w[None, ...]).mean(1)

        assert np.isfinite(suff_stats).all()
        return cls.from_sufficient_statistics(suff_stats, log_norm=log_norm)

    @classmethod
    def from_mode(cls, mode, covariance):
        pass

    def log_normalisation(self, *dists):
        """
        Calculates the log of the integral of the product of a
        set of distributions

        NOTE: ignores log normalisation
        """
        prod_dist = self.sum_natural_parameters(self, *self._iter_dists(dists))

        log_numerator = sum(
            dist.log_base_measure - dist.log_partition
            for dist in dists)
        log_denominator = prod_dist.log_base_measure - prod_dist.log_partition

        return log_numerator - log_denominator

    @staticmethod
    def _iter_dists(dists):
        for elem in dists:
            if isinstance(elem, AbstractMessage):
                yield elem
            elif np.isscalar(elem):
                yield elem
            else:
                for dist in elem:
                    yield dist

    def update_invalid(self, other: 'AbstractMessage') -> 'AbstractMessage':
        invalid = reduce(
            and_, (np.isfinite(p) for p in
                   chain(self.parameters, [self.check_support()])))
        if self.ndim:
            valid_parameters = (
                np.where(invalid, p, p_safe) for p, p_safe in zip(self, other))
        else:
            valid_parameters = self if invalid else other  # TODO: Fairly certain this would not work
        return type(self)(*valid_parameters, log_norm=self.log_norm)

    def check_support(self) -> np.ndarray:
        if self._parameter_support is not None:
            return reduce(
                and_,
                ((p >= support[0]) & (p <= support[1])
                 for p, support in zip(self.parameters, self._parameter_support)))
        elif self.ndim:
            return np.array(True, dtype=bool, ndmin=self.ndim)
        return np.array([True])

    @property
    def is_valid(self):
        return np.isfinite(
            self.natural_parameters
        ).all() and np.all(
            self.check_support()
        )

    @staticmethod
    def _get_mean_variance(mean: np.ndarray, covariance: float
                           ) -> Tuple[np.ndarray, np.ndarray]:
        mean, covariance = np.asanyarray(mean), np.asanyarray(covariance)

        if not covariance.shape:
            # If variance is float simply pass through
            variance = covariance * np.ones_like(mean)
            if not variance.shape:
                variance = variance.item()
        elif mean.shape == covariance.shape:
            variance = np.asanyarray(covariance)
        elif covariance.shape == mean.shape * 2:
            # If 2D covariance matrix passed get diagonal
            inds = tuple(np.indices(mean.shape))
            variance = np.asanyarray(covariance)[inds * 2]
        else:
            raise ValueError(
                f"shape of covariance {covariance.shape} is invalid "
                f"must be (), {mean.shape}, or {mean.shape * 2}")
        return mean, variance



import os
import shutil
from os import path

import pytest

import autofit as af
from autoconf import conf
from test_autofit import mock


directory = path.dirname(path.realpath(__file__))


@pytest.fixture(
    autouse=True
)
def remove_reports():
    yield
    for d, _, files in os.walk(directory):
        for file in files:
            if file == "report.log":
                os.remove(f"{d}/{file}")


@pytest.fixture(autouse=True)
def set_config_path():
    conf.instance = conf.Config(
        config_path=path.join(directory, "unit/config"),
        output_path=path.join(directory, "output")
    )


@pytest.fixture(autouse=True)
def remove_output():
    try:
        shutil.rmtree(f"{directory}/output")
    except FileNotFoundError:
        pass


@pytest.fixture
def model():
    return af.ModelMapper()


@pytest.fixture(name="phase")
def make_phase():
    phase = af.AbstractPhase(phase_name="phase name", search=af.MockSearch())
    phase.model.one = af.PriorModel(mock.MockComponents, component=mock.MockClassx2)
    return phase


@pytest.fixture(name="collection")
def make_collection():
    collection = af.ResultsCollection()
    model = af.ModelMapper()
    model.one = af.PriorModel(mock.MockComponents, component=mock.MockClassx2)
    instance = af.ModelInstance()
    instance.one = mock.MockComponents(component=mock.MockClassx2())

    result = af.MockResult(model=model, instance=instance)

    model = af.ModelMapper()
    instance = af.ModelInstance()

    model.hyper_galaxy = mock.HyperGalaxy
    instance.hyper_galaxy = mock.HyperGalaxy()

    hyper_result = af.MockResult(model=model, instance=instance)

    result.hyper_result = hyper_result

    collection.add("phase name", result)

    return collection



import copy
import inspect
import logging

from typing_inspect import is_tuple_type

from autofit.mapper.model_object import ModelObject
from autofit.mapper.prior import TuplePrior, Prior
from autofit.mapper.prior.deferred import DeferredInstance
from autofit.mapper.prior.promise import Promise
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.mapper.prior_model.abstract import check_assertions

logger = logging.getLogger(__name__)


class PriorModel(AbstractPriorModel):
    """Object comprising class and associated priors
        @DynamicAttrs
    """

    @property
    def name(self):
        return self.cls.__name__

    def as_model(self):
        return PriorModel(self.cls)

    def __hash__(self):
        return self.id

    def __add__(self, other):
        if self.cls != other.cls:
            raise TypeError(
                f"Cannot add PriorModels with different classes "
                f"({self.cls.__name__} and {other.cls.__name__})"
            )
        return super().__add__(other)

    @property
    def constructor_argument_names(self):
        try:
            return inspect.getfullargspec(self.cls).args[1:]
        except TypeError:
            return []

    def __init__(self, cls, **kwargs):
        """
        Parameters
        ----------
        cls: class
            The class associated with this instance
        """
        super().__init__()
        if cls is self:
            return

        self.cls = cls

        try:
            annotations = inspect.getfullargspec(cls).annotations
        except TypeError:
            annotations = dict()

        try:
            arg_spec = inspect.getfullargspec(cls)
            defaults = dict(
                zip(arg_spec.args[-len(arg_spec.defaults):], arg_spec.defaults)
            )
        except TypeError:
            defaults = {}

        args = self.constructor_argument_names

        if "settings" in defaults:
            del defaults["settings"]
        if "settings" in args:
            args.remove("settings")

        for arg in args:
            if isinstance(defaults.get(arg), str):
                continue

            if arg in kwargs:
                keyword_arg = kwargs[arg]
                if isinstance(keyword_arg, (list, dict)):
                    from autofit.mapper.prior_model.collection import (
                        CollectionPriorModel,
                    )

                    ls = CollectionPriorModel(keyword_arg)

                    setattr(self, arg, ls)
                else:
                    if inspect.isclass(keyword_arg):
                        keyword_arg = PriorModel(keyword_arg)
                    setattr(self, arg, keyword_arg)
            elif arg in defaults and isinstance(defaults[arg], tuple):
                tuple_prior = TuplePrior()
                for i in range(len(defaults[arg])):
                    attribute_name = "{}_{}".format(arg, i)
                    setattr(
                        tuple_prior, attribute_name, self.make_prior(attribute_name)
                    )
                setattr(self, arg, tuple_prior)
            elif arg in annotations and annotations[arg] != float:
                spec = annotations[arg]
                # noinspection PyUnresolvedReferences
                if inspect.isclass(spec) and issubclass(spec, float):
                    from autofit.mapper.prior_model.annotation import (
                        AnnotationPriorModel,
                    )

                    setattr(self, arg, AnnotationPriorModel(spec, cls, arg))
                elif is_tuple_type(spec):
                    tuple_prior = TuplePrior()
                    for i, tuple_arg in enumerate(spec.__args__):
                        attribute_name = "{}_{}".format(arg, i)
                        setattr(
                            tuple_prior, attribute_name, self.make_prior(attribute_name)
                        )
                    setattr(self, arg, tuple_prior)
                else:
                    setattr(self, arg, PriorModel(annotations[arg]))
            else:
                setattr(self, arg, self.make_prior(arg))
        for key, value in kwargs.items():
            if not hasattr(self, key):
                setattr(
                    self, key, PriorModel(value) if inspect.isclass(value) else value
                )

    def __eq__(self, other):
        return (
                isinstance(other, PriorModel)
                and self.cls == other.cls
                and self.prior_tuples == other.prior_tuples
        )

    def make_prior(self, attribute_name):
        """
        Create a prior for an attribute of a class with a given name. The prior is
        created by searching the default prior config for the attribute.

        Entries in configuration with a u become uniform priors; with a g become
        gaussian priors; with a c become instances.

        If prior configuration for a given attribute is not specified in the
        configuration for a class then the configuration corresponding to the parents
        of that class is searched. If no configuration can be found then a prior
        exception is raised.

        Parameters
        ----------
        attribute_name: str
            The name of the attribute for which a prior is created

        Returns
        -------
        prior: p.Prior
            A prior

        Raises
        ------
        exc.PriorException
            If no configuration can be found
        """
        cls = self.cls
        if not inspect.isclass(cls):
            # noinspection PyProtectedMember
            cls = inspect._findclass(cls)
        return Prior.for_class_and_attribute_name(cls, attribute_name)

    def __setattr__(self, key, value):
        if key not in (
                "component_number",
                "phase_property_position",
                "mapping_name",
                "id",
        ):
            try:
                if "_" in key:
                    name = key.split("_")[0]
                    tuple_prior = [v for k, v in self.tuple_prior_tuples if name == k][
                        0
                    ]
                    setattr(tuple_prior, key, value)
                    return
            except IndexError:
                pass
        try:
            super().__setattr__(key, value)
        except AttributeError as e:
            logger.exception(e)
            logger.exception(key)

    def __getattr__(self, item):
        try:
            if "_" in item:
                return getattr(
                    [v for k, v in self.tuple_prior_tuples if item.split("_")[0] == k][
                        0
                    ],
                    item,
                )

        except IndexError:
            pass
        self.__getattribute__(item)

    @property
    def is_deferred_arguments(self):
        return len(self.direct_deferred_tuples) > 0

    # noinspection PyUnresolvedReferences
    @check_assertions
    def _instance_for_arguments(self, arguments: {ModelObject: object}):
        """
        Create an instance of the associated class for a set of arguments

        Parameters
        ----------
        arguments: {Prior: float}
            Dictionary mapping_matrix priors to attribute analysis_path and value pairs

        Returns
        -------
            An instance of the class
        """
        model_arguments = dict()
        attribute_arguments = {
            key: value
            for key, value in self.__dict__.items()
            if key in self.constructor_argument_names
        }

        for tuple_prior in self.tuple_prior_tuples:
            model_arguments[tuple_prior.name] = tuple_prior.prior.value_for_arguments(
                arguments
            )
        for prior_model_tuple in self.direct_prior_model_tuples:
            prior_model = prior_model_tuple.prior_model
            model_arguments[
                prior_model_tuple.name
            ] = prior_model.instance_for_arguments(arguments)

        prior_arguments = dict()

        for name, prior in self.direct_prior_tuples:
            try:
                prior_arguments[name] = arguments[prior]
            except KeyError as e:
                raise KeyError(
                    f"No argument given for prior {name}"
                ) from e

        constructor_arguments = {
            **attribute_arguments,
            **model_arguments,
            **prior_arguments,
        }

        if self.is_deferred_arguments:
            return DeferredInstance(self.cls, constructor_arguments)

        if not inspect.isclass(self.cls):
            result = object.__new__(inspect._findclass(self.cls))
            cls = self.cls
            cls(result, **constructor_arguments)
        else:
            result = self.cls(**constructor_arguments)

        for key, value in self.__dict__.items():
            if (
                    not hasattr(result, key)
                    and not isinstance(value, Prior)
                    and not isinstance(value, Promise)
            ):
                if isinstance(value, PriorModel):
                    value = value.instance_for_arguments(arguments)
                elif isinstance(value, Prior):
                    value = arguments[value]
                try:
                    setattr(result, key, value)
                except AttributeError:
                    pass

        return result

    def gaussian_prior_model_for_arguments(self, arguments):
        """
        Create a new instance of model mapper with a set of Gaussian priors based on \
        tuples provided by a previous nonlinear search.

        Parameters
        ----------
        arguments: [(float, float)]
            Tuples providing the mean and sigma of gaussians

        Returns
        -------
        new_model: ModelMapper
            A new model mapper populated with Gaussian priors
        """
        new_model = copy.deepcopy(self)
        new_model._assertions = list()

        model_arguments = {t.name: arguments[t.prior] for t in self.direct_prior_tuples}

        for tuple_prior_tuple in self.tuple_prior_tuples:
            setattr(
                new_model,
                tuple_prior_tuple.name,
                tuple_prior_tuple.prior.gaussian_tuple_prior_for_arguments(arguments),
            )
        for prior_tuple in self.direct_prior_tuples:
            setattr(new_model, prior_tuple.name, model_arguments[prior_tuple.name])
        for instance_tuple in self.instance_tuples:
            setattr(new_model, instance_tuple.name, instance_tuple.instance)

        for name, prior_model in self.direct_prior_model_tuples:
            setattr(
                new_model,
                name,
                prior_model.gaussian_prior_model_for_arguments(arguments),
            )

        return new_model



import autofit as af

from howtofit.chapter_1_introduction.tutorial_5_visualization_masking.src.fit import (
    fit as f,
)
from howtofit.chapter_1_introduction.tutorial_5_visualization_masking.src.phase import (
    visualizer,
)

"""
The 'analysis.py' module in this tutorial has two changes from the previous tutorial:

   - The Analysis class is passed a mask, which is used to mask the data and thus fit only specific regions of it.
   - The Analysis class has a Visualizer, which performs visualization during and after the non-linear search.
"""


class Analysis(af.Analysis):
    def __init__(self, masked_dataset, image_path=None):

        super().__init__()

        """The masked-dataset is created in the 'phase.py' module, in the 'make_analysis' method."""

        self.masked_dataset = masked_dataset

        """
        The visualizer is the tool that we'll use the visualize a phase's unmasked dataset (before the model-fitting
        begins) and the best-fit solution found by the model-fit (during and after the model-fitting).

        Check out 'visualizer.py' for more details.
        """

        self.visualizer = visualizer.Visualizer(
            masked_dataset=self.masked_dataset, image_path=image_path
        )

    def log_likelihood_function(self, instance):
        """Determine the log likelihood of a fit of a Gaussian to the dataset, using the model instance of a Gaussian.

        Parameters
        ----------
        instance
            The Gaussian model instance.

        Returns
        -------
        fit : Fit.log_likelihood
            The log likelihood value indicating how well this model fit the masked dataset.
        """
        model_data = self.model_data_from_instance(instance=instance)
        fit = self.fit_from_model_data(model_data=model_data)
        return fit.log_likelihood

    def model_data_from_instance(self, instance):
        return instance.gaussian.profile_from_xvalues(
            xvalues=self.masked_dataset.xvalues
        )

    def fit_from_model_data(self, model_data):
        return f.FitDataset(masked_dataset=self.masked_dataset, model_data=model_data)

    def visualize(self, instance, during_analysis):
        """
        During a phase, the 'visualize' method is called throughout the non-linar search. The 'instance' passed into
        the visualize method is highest log likelihood solution obtained by the model-fit so far.

        In the analysis we use this instance to create the best-fit fit of our model-fit.
        """

        model_data = self.model_data_from_instance(instance=instance)
        fit = self.fit_from_model_data(model_data=model_data)

        """The visualizer now outputs images of the best-fit results to hard-disk (checkout 'visualizer.py')."""

        self.visualizer.visualize_fit(fit=fit, during_analysis=during_analysis)



from howtofit.chapter_1_introduction.tutorial_8_aggregator.src.plot import line_plots

"""The 'fit_plots.py' module is unchanged from the previous tutorial."""


def data(fit, output_path=None, output_filename=None, output_format="show"):
    """Plot the data values of a Fit.

    Parameters
    -----------
    Fit : fit.Fit
        The observed Fit dataset whose data is plotted.
    """
    line_plots.line(
        xvalues=fit.xvalues,
        line=fit.data,
        ylabel="Data Values",
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )


def noise_map(fit, output_path=None, output_filename=None, output_format="show"):
    """Plot the noise-map of a Fit.

    Parameters
    -----------
    Fit : fit.Fit
        The observed Fit whose noise-map is plotted.
    """
    line_plots.line(
        xvalues=fit.xvalues,
        line=fit.noise_map,
        ylabel="Noise Map",
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )


def signal_to_noise_map(
    fit, output_path=None, output_filename=None, output_format="show"
):
    """Plot the signal-to-noise-map of a Fit.

    Parameters
    -----------
    Fit : fit.Fit
        The observed Fit whose signal-to-noise-map is plotted.
    """
    line_plots.line(
        xvalues=fit.xvalues,
        line=fit.signal_to_noise_map,
        ylabel="Signal-To-Noise Map",
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )


def model_data(fit, output_path=None, output_filename=None, output_format="show"):
    """Plot the model data of a Fit.

    Parameters
    -----------
    Fit : fit.Fit
        The Fit model data.
    """
    line_plots.line(
        xvalues=fit.xvalues,
        line=fit.model_data,
        ylabel="Model Data",
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )


def residual_map(fit, output_path=None, output_filename=None, output_format="show"):
    """Plot the residual-map of a Fit.

    Parameters
    -----------
    Fit : fit.Fit
        The observed Fit whose residual-map is plotted.
    """
    line_plots.line(
        xvalues=fit.xvalues,
        line=fit.residual_map,
        ylabel="Residual Map",
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )


def normalized_residual_map(
    fit, output_path=None, output_filename=None, output_format="show"
):
    """Plot the normalized_residual-map of a Fit.

    Parameters
    -----------
    Fit : fit.Fit
        The observed Fit whose normalized residual-map is plotted.
    """
    line_plots.line(
        xvalues=fit.xvalues,
        line=fit.normalized_residual_map,
        ylabel="Normalized Residual Map",
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )


def chi_squared_map(fit, output_path=None, output_filename=None, output_format="show"):
    """Plot the chi_squared-map of a Fit.

    Parameters
    -----------
    Fit : fit.Fit
        The observed Fit whose chi-squared-map is plotted.
    """
    line_plots.line(
        xvalues=fit.xvalues,
        line=fit.chi_squared_map,
        ylabel="Chi-Squared Map",
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )



import numpy as np

"""The 'dataset.py' module has been extended to give the dataset a name and metadata."""


class Dataset:
    def __init__(self, data, noise_map, name=None):
        """A class containing the data and noise-map of a 1D line dataset.

        Parameters
        ----------
        data : np.ndarray
            The array of the data, in arbitrary units.
        noise_map : np.ndarray
            An array describing the RMS standard deviation error in each data pixel, in arbitrary units.
        """
        self.data = data
        self.noise_map = noise_map

        # The name of the dataset is used by the aggregator, to determine the name of the file the dataset is saved as
        # and so that when using the aggregator you can know which dataset you are manipulating.

        self.name = name if name is str else "dataset"

    @property
    def xvalues(self):
        return np.arange(self.data.shape[0])


class MaskedDataset:
    def __init__(self, dataset, mask):
        """
        A masked dataset, which is an image, noise-map and mask.

        Parameters
        ----------
        dataset: im.Dataset
            The dataset (the image, noise-map, etc.)
        mask: msk.Mask
            The 1D mask that is applied to the dataset.
        """

        self.dataset = dataset
        self.mask = mask
        self.data = dataset.data * np.invert(mask)
        self.noise_map = dataset.noise_map * np.invert(mask)

    @property
    def xvalues(self):
        return np.arange(self.data.shape[0])

    def signal_to_noise_map(self):
        return self.data / self.noise_map

    def with_left_trimmed(self, data_trim_left):

        if data_trim_left is None:
            return self

        """Here, we use the existing masked dataset to create a trimmed dataset."""

        data_trimmed = self.dataset.data[data_trim_left:]
        noise_map_trimmed = self.dataset.noise_map[data_trim_left:]

        dataset_trimmed = Dataset(data=data_trimmed, noise_map=noise_map_trimmed)

        mask_trimmed = self.mask[data_trim_left:]

        return MaskedDataset(dataset=dataset_trimmed, mask=mask_trimmed)

    def with_right_trimmed(self, data_trim_right):

        if data_trim_right is None:
            return self

        """We do the same as above, but removing data to the right."""

        data_trimmed = self.dataset.data[:-data_trim_right]
        noise_map_trimmed = self.dataset.noise_map[:-data_trim_right]

        dataset_trimmed = Dataset(data=data_trimmed, noise_map=noise_map_trimmed)

        mask_trimmed = self.mask[:-data_trim_right]

        return MaskedDataset(dataset=dataset_trimmed, mask=mask_trimmed)
